#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 14.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Union, Iterable, Dict, Tuple

# External modules
import xarray as xr
from xarray import register_dataarray_accessor

# Internal modules


logger = logging.getLogger(__name__)


class StateError(Exception):  # pragma: no cover
    """
    This error is an error if given state is not valid or if there is
    something strange with the state.
    """
    pass


@register_dataarray_accessor('state')
class ModelState(object):
    """
    This accessor extends an :py:class:`xarray.DataArray` by a ``state``
    property. This state property is used as utility class for an easier
    processing of model states for data assimilation purpose.

    Arguments
    ---------
    xr_da : :py:class:`xarray.DataArray`
        This model state accessor is registered to this array. The array should
        have (``variable``, ``time``, ``ensemble``, ``grid``) as coordinates,
        which are explained below. The values of the array have to be float.

        Coordinate explanation:
            ``var_name`` – str
                This coordinate allows to concatenate multiple variables into
                one single array. The coordinate should have :py:class:`str` as
                dtype.

            ``time`` – datetime like
                Different times can be concatenated into this coordinate. This
                coordinate is useful for data assimilation algorithms which
                allow assimilation of time dependent variables. The coordinate
                should be a datetime like dtype.

            ``ensemble`` – int
                This coordinate is used for ensemble based data assimilation
                algorithms. The ensemble members are numbered as
                :py:class:`int`. An integer value of 0 symbolizes a
                deterministic or control run. In some ensemble based algorithms,
                the ensemble coordinate is looped.

            ``grid``
                This coordinate is used to characterize the spatial position
                of one single value within the array. This coordinate can be a
                :py:class:`~pandas.MultiIndex`, where different coordinate
                components are stacked. In some algorithms the analysis is
                calculated for every grid point independently such that the
                algorithm loops over this coordinate.
    """
    def __init__(self, xr_da: xr.DataArray):
        self.array = xr_da

    def __str__(self):
        return 'ModelState({0}'.format(str(self.array))

    def __repr__(self):
        return 'ModelState'

    @property
    def _valid_dims(self) -> bool:
        """
        Checks if the dimension of this array are valid. The dimensions need
        to be in right order and have the right names as specified in
        documentation.

        Returns
        -------
        valid_dims : bool
            If the dimensions of this array have the right name and right order.
        """
        correct_dims = ('var_name', 'time', 'ensemble', 'grid')
        valid_dims = correct_dims == self.array.dims
        return valid_dims

    @property
    def valid(self) -> bool:
        """
        Checks if the array has the right form, coordinates and dimensions.

        Returns
        -------
        valid_array : bool
            If the given array is valid.
        """
        valid_array = self._valid_dims
        return valid_array

    def split_mean_perts(self,
                         dim: Union[str, Iterable[str]] = 'ensemble',
                         axis: Union[int, Iterable[int]] = None,
                         **kwargs: Dict) -> Tuple[xr.DataArray, xr.DataArray]:
        """
        Splits this :py:class:`~xarray.DataArray` into a mean array and a
        perturbations array by given dimension.

        Parameters
        ----------
        dim : str or sequence of str, optional
            The array is split over this dimension(s). Default is `'ensemble'`.
        axis : int or sequence of int, optional
            The array is split over this axis/axes. Only ``dim`` or ``axis`` can
            be supplied. Default is None.
        **kwargs : dict
            These additional keyword arguments are passed onto
            :py:meth:`~xarray.DataArray.mean` method of set array.

        Returns
        -------
        mean : :py:class:`~xarray.DataArray`
            Newly created array with averaged values. This array has the same
            coordinates as the original array.
        perts : :py:class:`~xarray.DataArray`
            Newly created array with perturbations as values. The perturbations
            are the difference of ``original array - mean``.
            This array has the same coordinates as the original array.
        """
        mean = self.array.mean(dim=dim, axis=axis, **kwargs)
        perts = self.array - mean
        return mean, perts



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/11/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Union, List, Any, Dict

# External modules
from scipy.spatial import cKDTree
import numpy as np
import pandas as pd
import xarray as xr

# Internal modules
from ..base_ops import BaseOperator


logger = logging.getLogger(__name__)


EARTH_RADIUS = 6371000


class CosmoT2mOperator(BaseOperator):
    def __init__(
            self,
            station_df: pd.DataFrame,
            cosmo_coords: np.ndarray,
            cosmo_const: xr.Dataset,):
        """
        This 2-metre-temperature observation operator is used as observation
        operator for COSMO data. This observation operator selects the nearest
        grid point to given observations and corrects the height difference
        between COSMO and station height as written in the user guide of COSMO.

        Parameters
        ----------
        station_df : :py:class:`pandas.DataFrame`
            This station dataframe is used to determine the station position and
            height.
        cosmo_coords : :py:class:`numpy.ndarray`
            The cosmo coordinates as numpy array. The first axis should be the
            number of grid points, while the second axis is ('lat', 'lon').
        cosmo_const : :py:class:`xarray.Dataset`
            The constant data file of cosmo. `HHL` within the constant data is
            used to to estimate the lapse rate. `HSURF` within the constant data
            is used correct for station height. If no `HHL` is available,
            :py:meth:`~pytassim.obs_ops.terrsysmp.cos_t2m.CosmoT2mOperator.
            get_lapse_rate` has to be overwritten.
        """
        self._h_diff = None
        self._locs = None
        self.station_df = station_df
        self.cosmo_coords = cosmo_coords
        self.cosmo_const = cosmo_const
        self.lev_inds = [40, 35]

    def __str__(self) -> str:
        return 'COSMO 2-metre-temperature observation operator'

    def __repr__(self) -> str:
        return 'T2mOperator'

    @property
    def locs(self) -> np.ndarray:
        if self._locs is None:
            self._locs = self._calc_locs()
        return self._locs

    @property
    def height_diff(self) -> np.ndarray:
        if self._h_diff is None:
            self._h_diff = self._calc_h_diff()
        return self._h_diff

    @property
    def cosmo_height(self) -> np.ndarray:
        cosmo_hsurf = self.cosmo_const['HSURF'].stack(grid=['rlat', 'rlon'])
        cosmo_loc = self._localize_grid(cosmo_hsurf)
        cosmo_height = cosmo_loc.isel(time=0).values
        return cosmo_height

    @staticmethod
    def _get_cartesian(latlonalt) -> np.ndarray:
        lat_rad = np.deg2rad(latlonalt[:, 0])
        lon_rad = np.deg2rad(latlonalt[:, 1])
        x = EARTH_RADIUS * np.cos(lat_rad) * np.cos(lon_rad)
        y = EARTH_RADIUS * np.cos(lat_rad) * np.sin(lon_rad)
        z = EARTH_RADIUS * np.sin(lat_rad) + latlonalt[:, 2]
        xyz = np.stack([x, y, z], axis=-1)
        return xyz

    def _calc_locs(self) -> np.ndarray:
        station_lat_lon = self.station_df[['Breite', 'Länge']].values
        station_alt = self.station_df['Stations-\r\nhöhe'].values.reshape(-1, 1)
        station_llalt = np.concatenate([station_lat_lon, station_alt], axis=-1)
        station_xyz = self._get_cartesian(station_llalt)

        cosmo_lat_lon = self.cosmo_coords.reshape(-1, 2)
        cosmo_alt = self.cosmo_const['HSURF'].isel(time=0).values
        cosmo_alt = cosmo_alt.reshape(-1, 1)
        cosmo_llalt = np.concatenate([cosmo_lat_lon, cosmo_alt], axis=-1)
        cosmo_xyz = self._get_cartesian(cosmo_llalt)

        locs = self._get_neighbors(cosmo_xyz, station_xyz)
        locs = np.unravel_index(locs, self.cosmo_coords.shape[:2])
        return locs

    def _calc_h_diff(self) -> np.ndarray:
        station_height = self.station_df['Stations-\r\nhöhe'].values
        height_diff = station_height - self.cosmo_height
        return height_diff

    @staticmethod
    def _get_neighbors(
            src_points: np.ndarray,
            trg_points: np.ndarray
    ) -> np.ndarray:
        tree = cKDTree(src_points)
        _, locs = tree.query(trg_points, k=1)
        return locs

    def _localize_grid(
            self,
            ds: xr.DataArray,
            height_ind: Union[None, int] = None,
            height_lev: Union[None, float] = None
    ) -> xr.DataArray:
        grid_ind = ds.indexes['grid']
        rlat = grid_ind.levels[0][self.locs[0]].values
        rlon = grid_ind.levels[1][self.locs[1]].values
        if 'vgrid' in grid_ind.names and height_lev is not None:
            height = [height_lev] * len(rlat)
            loc_list = list(zip(rlat, rlon, height))
        elif 'vgrid' in grid_ind.names and height_ind is not None:
            height = [grid_ind.levels[2][height_ind], ] * len(rlat)
            loc_list = list(zip(rlat, rlon, height))
        elif 'vgrid' not in grid_ind.names:
            loc_list = list(zip(rlat, rlon))
        else:
            raise ValueError('An height index has to be given to localize!')
        localized_ds = ds.sel(grid=loc_list)
        return localized_ds

    def get_lapse_rate(self, cosmo_ds: xr.DataArray) -> xr.DataArray:
        height = cosmo_ds.indexes['grid'].get_level_values('vgrid')
        h_diff = height[self.lev_inds[1]] - height[self.lev_inds[0]]

        sel_temp = cosmo_ds.sel(var_name='T')
        temp_1 = self._localize_grid(sel_temp, height_ind=self.lev_inds[1])
        temp_0 = self._localize_grid(sel_temp, height_ind=self.lev_inds[0])
        temp_diff = temp_1.values - temp_0.values

        lapse_rate = temp_diff / h_diff
        return lapse_rate

    def obs_op(
            self,
            in_array: xr.DataArray,
            *args: List[Any],
            **kwargs: Dict[str, Any]
    ) -> xr.DataArray:
        uncorr_t2m = self._localize_grid(in_array.sel(var_name='T_2M'),
                                         height_lev=0)
        correction = self.height_diff * self.get_lapse_rate(in_array)
        corr_t2m = uncorr_t2m + correction
        return corr_t2m



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 14.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc
import warnings
import time
import datetime
from typing import Union, Iterable, Tuple, Any, List

# External modules
import xarray as xr
import pandas as pd
import torch
import numpy as np

# Internal modules
from .utils import grid_to_array

from pytassim.state import StateError
from pytassim.observation import ObservationError
from pytassim.transform import BaseTransformer


logger = logging.getLogger(__name__)


class BaseAssimilation(object):
    """
    BaseAssimilation is used as base class for all assimilation algorithms for
    fast prototyping of assimilation prototyping. To implement a data
    assimilation, one needs to overwrite
    :py:meth:`~pytassim.assimilation.base.BaseAssimilation.update_state`.
    """
    def __init__(self, smoother: bool = False, gpu: bool = False,
                 pre_transform: Union[None, Iterable[BaseTransformer]] = None,
                 post_transform: Union[None, Iterable[BaseTransformer]] = None):
        self.smoother = smoother
        self.gpu = gpu
        self.pre_transform = pre_transform
        self.post_transform = post_transform
        self.dtype = torch.double

    def __str__(self):
        return 'BaseAssimilation'

    def __repr__(self):
        return 'BaseAssimilation'

    def _states_to_torch(
            self,
            *states: Tuple[np.ndarray]
    ) -> Tuple[torch.Tensor]:
        torch_states = [torch.from_numpy(s).to(self.dtype) for s in states]
        if self.gpu:
            torch_states = [s.cuda() for s in torch_states]
        return torch_states

    @staticmethod
    def _validate_state(state: xr.DataArray):
        if not isinstance(state, xr.DataArray):
            raise TypeError('*** Given state is not a valid '
                            '``xarray.DataArray`` ***\n{0:s}'.format(state))
        if not state.state.valid:
            err_msg = '*** Given state is not a valid state ***\n{0:s}'
            raise StateError(err_msg.format(str(state)))

    def _validate_single_obs(self, observation: xr.Dataset):
        if not isinstance(observation, xr.Dataset):
            raise TypeError('*** Given observation is not a valid'
                            '``xarray.Dataset`` ***\n{0:s}'.format(observation))
        if not observation.obs.valid:
            err_msg = '*** Given observation is not a valid observation ***' \
                      '\n{0:s}'
            raise ObservationError(err_msg.format(str(observation)))
        if observation.obs.correlated != self._correlated:
            err_msg = '*** The correlation of the observation {0:s} is not ' \
                      'the same as the asked correlation of this ' \
                      'assimilation algorithm (asked: {1:s}, actual: {2:s}) ***'
            raise ObservationError(
                err_msg.format(str(observation), str(self._correlated),
                               str(observation.obs.correlated))
            )

    def _validate_observations(
            self, observations: Union[xr.Dataset, Iterable[xr.Dataset]]
    ):
        if isinstance(observations, (list, set, tuple)):
            for obs in observations:
                self._validate_single_obs(obs)
        else:
            self._validate_single_obs(observations)

    @staticmethod
    def _get_analysis_time(
            state: xr.Dataset,
            analysis_time: Any = None
    ) -> pd.Timestamp:
        if analysis_time is None:
            valid_time = state.time[-1]
        else:
            analysis_time = pd.to_datetime(analysis_time)
            try:
                valid_time = state.time.sel(time=analysis_time, method=None)
            except KeyError:
                valid_time = state.time.sel(time=analysis_time,
                                            method='nearest')
                warnings.warn(
                    'Given analysis time {0:s} is not within state, used '
                    'instead nearest neighbor {1:s}'.format(
                        str(analysis_time), str(valid_time.values)
                    ),
                    category=UserWarning
                )
        valid_time = valid_time.values
        valid_time = pd.to_datetime(valid_time)
        return valid_time

    @staticmethod
    def _apply_obs_operator(
            pseudo_state: xr.DataArray,
            observations: Iterable[xr.Dataset]
    ) -> Tuple[List[xr.DataArray], List[xr.Dataset]]:
        """
        This method applies the observation operator on given state. The
        observation operator has to be set within given observations. It is
        possible to overwrite this method to implement an own observation
        operator, which was not set in given observations.

        Parameters
        ----------
        pseudo_state : :py:class:`xarray.DataArray`
            This state is used as base state to apply set observation operator.
        observations : iterable(:py:class:`xarray.Dataset`)
            These observations are used as basis for the observation operators.
            The observation operator should be set as method
            :py:meth:`xarray.Dataset.obs.operator`.

        Returns
        -------
        obs_equivalent : list(:py:class:`xarray.DataArray`)
            A list with observation equivalents as :py:class:`xarray.DataArray`.
            These observation equivalents have three dimensions, ``ensemble``,
            ``time`` and ``obs_grid``. The order within these observation
            equivalent is the same as in the observations.
        filtered_observations : list(:py:class:`xarray.Dataset`)
            These observations are filtered such that observations without an
            observation operator are dropped.
        """
        obs_equivalent = []
        filtered_observations = []
        for obs in observations:
            try:
                obs_equivalent.append(obs.obs.operator(obs, pseudo_state))
                filtered_observations.append(obs)
            except NotImplementedError:
                pass
        return obs_equivalent, filtered_observations

    @abc.abstractmethod
    def _get_obs_cov(self, observations: Iterable[xr.Dataset]) -> np.ndarray:
        pass

    def _prepare_obs(
            self,
            observations: List[xr.Dataset]
    ) -> Tuple[np.ndarray, np.ndarray]:
        state_stacked_list = []
        for obs in observations:
            if isinstance(obs.indexes['obs_grid_1'], pd.MultiIndex):
                obs['obs_grid_1'] = pd.Index(
                    obs.indexes['obs_grid_1'].values, tupleize_cols=False
                )
            stacked_obs = obs['observations'].stack(
                obs_id=('time', 'obs_grid_1')
            )
            state_stacked_list.append(stacked_obs)
        state_concat = xr.concat(state_stacked_list, dim='obs_id')
        state_values = state_concat.values
        state_grid = grid_to_array(state_concat['obs_grid_1'].values)
        return state_values, state_grid

    @abc.abstractmethod
    def update_state(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]],
            pseudo_state: xr.DataArray,
            analysis_time: pd.Timestamp
    ) -> xr.DataArray:
        """
        This method is called by
        :py:meth:`~pytassim.assimilation.base.BaseAssimilation.assimilate` and
        has to be overwritten to implement a data assimilation algorithm.

        Parameters
        ----------
        state : :py:class:`xarray.DataArray`
            This state is updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.
            The :py:class:`xarray.Dataset` are validated with
            :py:class:`pytassim.observation.Observation.valid`
        pseudo_state : :py:class:`xarray.DataArray`
            This state is used to generate an observation-equivalent. This
             :py:class:`~xarray.DataArray` should have four coordinates, which
             are specified in :py:class:`pytassim.state.ModelState`.
        analysis_time : :py:class:`datetime.datetime`
            This analysis time determines at which point the state is updated.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            The analysed state based on given state and observations. The
            analysis has same coordinates as given ``state`` except ``time``,
            which should contain only one time step.
        """
        pass

    def assimilate(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]],
            pseudo_state: Union[xr.DataArray, None] = None,
            analysis_time: Any = None
    ) -> xr.DataArray:
        """
        This assimilate the ``observations`` in given background ``state`` and
        creates an analysis for given ``analysis_time``. The observations need
        an observation operator, which translate given state into an
        observation-equivalent. The state, observations, observation covariance
        and and the observation-equivalent are used to update given state. This
        method validates given state and observations, gets analysis time and
        calls
        :py:meth:`~pytassim.assimilation.base.BaseAssimilation.update_state`.
        If state to assimilate is given, this state is translated into
        observation space.

        Parameters
        ----------
        state : :py:class:`xarray.DataArray`
            This state is updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`. If no pseudo_state is
            specified, this state is also used to generate pseudo observations.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.
            The :py:class:`xarray.Dataset` are validated with
            :py:class:`pytassim.observation.Observation.valid`
        pseudo_state : :py:class:`xarray.DataArray` or None
            If this additional state is given, this state is used to create
            pseudo-observations. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        analysis_time : :py:class:`datetime.datetime` or None, optional
            This analysis time determines at which point the state is updated.
            If the analysis time is None, than the last time point in given
            state is used.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            The analysed state based on given state and observations. The
            analysis has same coordinates as given ``state`` except ``time``,
            which contains only one time step.
        """
        start_time = time.time()
        logger.info('Starting assimilation')
        if not observations:
            warnings.warn('No observation is given, I will return the '
                          'background state!', UserWarning)
            return state
        if not isinstance(observations, (list, set, tuple)):
            observations = (observations, )
        if pseudo_state is None:
            pseudo_state = state
        self._validate_state(state)
        self._validate_state(pseudo_state)
        self._validate_observations(observations)
        analysis_time = self._get_analysis_time(state, analysis_time)
        if isinstance(analysis_time, datetime.datetime):
            logger.info(
                'Analysis time: {0:s}'.format(
                    analysis_time.strftime('%Y-%m-%d %H:%M UTC')
                )
            )
        if self.smoother:
            back_state = state
        else:
            logger.info('Assimilation in non-smoother mode')
            pseudo_state = pseudo_state.sel(time=[analysis_time, ])
            back_state = state.sel(time=[analysis_time, ])
            sel_obs = []
            for obs in observations:
                tmp_obs = obs.sel(time=[analysis_time, ])
                tmp_obs.obs.operator = obs.obs.operator
                sel_obs.append(tmp_obs)
            observations = sel_obs
        if self.pre_transform:
            for trans in self.pre_transform:
                back_state, observations, pseudo_state = trans.pre(
                    back_state, observations, pseudo_state
                )
        logger.info('Finished with general preparation')
        analysis = self.update_state(back_state, observations, pseudo_state,
                                     analysis_time)
        logger.info('Created the analysis, starting with post-processing')
        if self.post_transform:
            for trans in self.post_transform:
                analysis = trans.post(analysis, back_state, observations,
                                      pseudo_state)
        self._validate_state(analysis)
        end_time = time.time()
        logger.info('Finished assimilation after {0:.2f} s'.format(
            end_time-start_time
        ))
        return analysis



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/12/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import types
import copy

# External modules

# Internal modules
from .base import BaseTransformer


logger = logging.getLogger(__name__)


class Normalizer(BaseTransformer):
    """
    This Normalizer normalize the first guess and observations before
    assimilation and reverts the normalization of the analysis after the
    assimilation. These methods can be used to generate normalized in- and
    outputs for machine learning algorithms. To normalize the mean value
    :math:`\\mu`` is subtracted from given field :math:`x`, which is then
    divided by given standard deviation :math:`\\sigma`,

    .. math::

       x_{\\text{norm}} = \\frac{x - \\mu}{\\sigma}.

    For post-processing this formula is reverted.



    Parameters
    ----------
    ens_stat : iterable(any)
        These ensemble statistics are used to normalize the background and to
        revert the normalization of the analysis. The first item of this
        iterable is used as mean value, while the second item is used as
        standard deviation.
    obs_stat : iterable(any)
        These observation statistics are used to normalize the observations
        before assimilation. This iterable should have at least the same length
        as given observations and is used in the same order as given
        observations. Every item should be an iterable, which is used for mean
        (first item) and standard deviation (second item).
    fg_stat : iterable(any)
        These first guess statistics are used to normalize the first guess
        array. The first item of this iterable is used as mean value, while the
        second item is used as standard deviation.
    """
    def __init__(self, ens_stat, obs_stat, fg_stat):
        self.ens_stat = ens_stat
        self.obs_stat = obs_stat
        self.fg_stat = fg_stat

    def pre(self, background, observations, first_guess):
        """
        This method normalizes given `background`, `observations` and
        `first_guess`by their set statistics with

        .. math::

           x_{\\text{norm}} = \\frac{x - \\mu}{\\sigma}.


        Parameters
        ----------
        background : :py:class:`xarray.DataArray`
            This array is normalized by set ensemble statistics. This first
            guess should be a valid state.
        observations : iterable(:py:class:`xarray.Dataset`)
            These observations are normalized by set observation statistics. To
            normalize these observations, the `observation` values are centered
            and scaled. These observations must be in the same order as set
            observation statistics. These observations should be valid
            observations with `observation` and `covariance` as
            :py:class:`xarray.DataArray`.
        first_guess : :py:class:`xarray.DataArray`
            This array is normalized by set ensemble statistics. This first
            guess should be a valid state.

        Returns
        -------
        background : :py:class:`xarray.DataArray`
            The centered and scaled background field.
        observations : iterable(:py:class:`xarray.Dataset`)
            The normalized observations with centered and scaled observation
            values.
        first_guess : :py:class:`xarray.DataArray`
            The centered and scaled first guess field.
        """
        background = (background - self.ens_stat[0]) / self.ens_stat[1]
        first_guess = (first_guess - self.fg_stat[0]) / self.fg_stat[1]
        obs_list = []
        for k, obs in enumerate(observations):
            tmp_obs = obs.copy(deep=True)
            tmp_obs['observations'] -= self.obs_stat[k][0]
            tmp_obs['observations'] /= self.obs_stat[k][1]
            tmp_obs.obs.operator = obs.obs.operator
            obs_list.append(tmp_obs)
        return background, obs_list, first_guess

    def post(self, analysis, background, observations, first_guess):
        """
        This method reverts the normalization of the analysis based on set
        ensemble statistics.

        Parameters
        ----------
        analysis : :py:class:`xarray.DataArray`
            The normalization of this analysis field is reverted. This should be
            a valid state.
        background : :py:class:`xarray.DataArray`
            This background is not used in this normalization method.
        observations : iterable(:py:class:`xarray.Dataset`)
            These observations are not used in this normalization method.
        first_guess : :py:class:`xarray.DataArray`
            This first guess is not used in this normalization method.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            This analysis was amplified by set standard deviation and shifted by
            set mean. This is a valid state.
        """
        analysis = analysis * self.ens_stat[1] + self.ens_stat[0]
        return analysis



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 11/29/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc
from typing import Callable, Any

# External modules

# Internal modules


logger = logging.getLogger(__name__)


class BaseIntegrator(object):
    """
    This is a base class for integrators. An integrators integrates given model
    function forward or backward in time, depending on given timestep.

    Arguments
    ---------
    model : func
        This model function takes a state and returns a new estimated state. The
        returned state should have the same shape as the input state. The model
        should be a time derivative such that it can be integrated. It is
        assumed that the state does not depend on the time itself.
    dt : float, optional
        This is the integration time step. This time step is unitless and
        depends on model's time unit. A positive time step indicates forward
        integration, while a negative shows a backward integration, which might
        be complicated for given model. Default is 0.05.
    """

    def __init__(self, model: Callable, dt: float = 0.05):
        self._model = None
        self._dt = None
        self.model = model
        self.dt = dt

    @property
    def model(self) -> Callable:
        """
        This model function takes a state and returns a new estimated state. The
        returned state should have the same shape as the input state. The model
        should be a time derivative such that it can be integrated. It is
        assumed that the state does not depend on the time itself.
        """
        return self._model

    @model.setter
    def model(self, new_model: Callable):
        if callable(new_model):
            self._model = new_model
        else:
            raise TypeError('Given model is not callable!')

    @property
    def dt(self) -> float:
        """
        This integration time step specifies the step width for the integration
        and is unit less, and depends on model's time unit. A positive time step
        indicates forward integration, while a negative shows a backward
        integration, which might be complicated for given model.
        """
        return self._dt

    @dt.setter
    def dt(self, new_dt: float):
        if not isinstance(new_dt, (float, int)):
            raise TypeError('Given time step is not a float!')
        elif new_dt == 0:
            raise ValueError('Given time step is zero!')
        else:
            self._dt = new_dt

    @abc.abstractmethod
    def _calc_increment(self, state: Any) -> Any:
        """
        This method estimates the increment based on given state, set model and
        time step.

        Parameters
        ----------
        state : any
            This state is used to estimate the increment.

        Returns
        -------
        est_inc : any
            This increment is estimated by this integration object.
        """
        pass

    def integrate(self, state: Any) -> Any:
        """
        This method integrates given model by set time step. Given state is used
        as initial state and passed to model.

        Parameters
        ----------
        state : any
            This state is used as initial state for the integration. This state
            is passed to set model.

        Returns
        -------
        int_state : any
            This state is integrated by given model. The integrated state is the
            initial state plus an increment estimated based on this integrator
            and set model.
        """
        estimated_inc = self._calc_increment(state)
        int_state = state + estimated_inc
        return int_state



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 11/29/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Callable, Any

# External modules

# Internal modules
from .integrator import BaseIntegrator


logger = logging.getLogger(__name__)


class RK4Integrator(BaseIntegrator):
    """
    RK4Integrator uses a Runge-Kutta fourth-order method to integrate given
    model function in time. This method calculates four different states, which
    are then averaged to one single state increment. To get another type of
    Runge-Kutta scheme, ``steps`` and ``weights`` can be changed.

    Arguments
    ---------
    model : func
        This model function takes a state and returns a new estimated state. The
        returned state should have the same shape as the input state. The model
        should be a time derivative such that it can be integrated. It is
        assumed that the state does not depend on the time itself.
    dt : float, optional
        This is the integration time step. This time step is unitless and
        depends on model's time unit. A positive time step indicates forward
        integration, while a negative shows a backward integration, which might
        be complicated for given model. Default is 0.05.
    """
    def __init__(self, model: Callable, dt: float = 0.05):
        super().__init__(model=model, dt=dt)
        self.steps = [0, self.dt / 2, self.dt / 2, self.dt]
        self.weights = [1, 2, 2, 1]
        self._weights_sum = sum(self.weights)
        self._weights = [w / self._weights_sum for w in self.weights]

    def __str__(self) -> str:
        return 'RK4Integrator(model={0:s}, dt={1})'.format(str(self.model),
                                                           self.dt)

    def __repr__(self) -> str:
        return 'RK4({0:s})'.format(repr(self.model))

    def _calc_increment(self, state: Any) -> Any:
        """
        This method estimates the increment based estimated slope and set time
        step.

        Parameters
        ----------
        state : any
            This state is used to estimate the slope.

        Returns
        -------
        est_inc : any
            This increment is estimated by multiplying estimated slope with
            set time step.
        """
        est_inc = self._estimate_slope(state) * self.dt
        return est_inc

    def _estimate_slope(self, state: Any) -> Any:
        """
        This method estimates the slope based on given state. This slope is
        used to calculate the increment.

        Parameters
        ----------
        state : any
            This state is used as initial state to estimates the slopes.

        Returns
        -------
        averaged_slope : any
            This slope is averaged based on estimated slopes. These slopes are
            calculated based on given state.
        """
        averaged_slope = state * 0
        curr_slope = state * 0
        for k, ts in enumerate(self.steps):
            model_state = state + curr_slope * ts
            curr_slope = self.model(model_state)
            averaged_slope += self._weights[k] * curr_slope
        return averaged_slope



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch
import torch.nn

# Internal modules
from .base_kernels import BaseKernel
from .utils import euclidean_dist


logger = logging.getLogger(__name__)


__all__ = ['GaussKernel', 'RBFKernel']


class GaussKernel(BaseKernel):
    """
    The Gaussian kernel is a type of radial basis function
    kernel :py:class:`pytassim.kernels.rbf.RBFKernel` and is implemented
    with a given lengthscale :math:`l`,

    .. math::

       K(x_i, x_j) = \\exp(-\\frac{{(x_i-x_j)}^2}{2\\,l^2}).


    Parameters
    ----------
    lengthscale : torch.Tensor, optional
        This lengthscale is used to estimate the Gaussian kernel. The
        default value of 1 assumes that the input is already normalized.
    """

    def __init__(self, lengthscale: torch.Tensor = torch.tensor(1.)):
        super().__init__()
        self.lengthscale = lengthscale

    def __str__(self):
        return "GaussKernel(l={0})".format(self.lengthscale)

    def __repr__(self):
        return "GaussKernel"

    def _get_lengthscale(self) -> torch.Tensor:
        return self.lengthscale

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x_scaled = x.div(self._get_lengthscale())
        y_scaled = y.div(self._get_lengthscale())
        euc_dist = euclidean_dist(x_scaled, y_scaled)
        factor = euc_dist / 2.
        k_mat = torch.exp(-factor)
        return k_mat


class RBFKernel(GaussKernel):
    """
    This radial basis function kernel is a universal kernel and is
    dependent on the chosen `gamma` :math:`\\gamma` factor,

    .. math::

       K(x_i, x_j) = \\exp(-\\gamma\\,{(x_i-x_j)}^2).

    Parameters
    ----------
    gamma : torch.Tensor, optional
        This gamma value if  is used to estimate the Gaussian kernel. The
        default value of 0.5 assumes that the input is already normalized.

    """
    def __init__(self, gamma: torch.Tensor = torch.tensor(0.5)):
        super().__init__()
        self.gamma = gamma

    def __str__(self):
        return "RBFKernel(\u03B3={0})".format(self.gamma)

    def __repr__(self):
        return "RBFKernel"

    def _get_lengthscale(self) -> torch.Tensor:
        return (0.5 / self.gamma) ** 0.5



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/14/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import operator
from typing import Union, Type, Iterable

# External modules
import numpy as np
import torch
import xarray as xr
import pandas as pd

import dask
import dask.array as da
from distributed import Client
from distributed.deploy.cluster import Cluster

# Internal modules
from .etkf_core import CorrMixin, UnCorrMixin
from .letkf import LETKFBase

from pytassim.localization import BaseLocalization
from pytassim.transform import BaseTransformer


logger = logging.getLogger(__name__)


class DistributedLETKFBase(LETKFBase):
    """
    Base object for a distributed localised ensemble transform Kalman filter.
    """
    def __init__(
            self,
            client: Union[None, Client] = None,
            cluster: Union[None, Type[Cluster]] = None,
            chunksize: int = 10,
            localization: Union[None, BaseLocalization] = None,
            inf_factor: Union[torch.Tensor, float, torch.nn.Parameter] = 1.0,
            smoother: bool = False, gpu: bool = False,
            pre_transform: Union[None, Iterable[Type[BaseTransformer]]] = None,
            post_transform: Union[None, Iterable[Type[BaseTransformer]]] = None
    ):
        super().__init__(localization, inf_factor, smoother, gpu, pre_transform,
                         post_transform)
        self._name = 'Distributed LETKF'
        self._cluster = None
        self._client = None
        self._chunksize = 1
        self.chunksize = chunksize
        self.set_client_cluster(client=client, cluster=cluster)

    @staticmethod
    def _validate_client(client) -> bool:
        return isinstance(client, Client)

    @staticmethod
    def _validate_cluster(cluster) -> bool:
        return hasattr(cluster, "scheduler_address")

    def _check_client_cluster(
            self,
            client: Union[None, Client],
            cluster: Union[None, Type[Cluster]],
    ):
        not_valid_cluster = not self._validate_cluster(cluster)
        not_valid_client = not self._validate_client(client)
        if not_valid_client and not_valid_cluster:
            raise ValueError(
                'Either a client or a cluster have to be specified!'
            )

    @property
    def cluster(self) -> Type[Cluster]:
        return self._cluster

    @property
    def client(self) -> Client:
        return self._client

    def set_client_cluster(
            self,
            client: Union[None, Client] = None,
            cluster: Union[None, Type[Cluster]] = None,
    ):
        """
        This method sets the client and cluster. If both are given and valid,
        then client has priority.

        Parameters
        ----------
        client : :py:class:``~dask.distributed.Client`` or None
            This dask distributed client is used to parallelize the processes.
            Either this client or ``cluster`` has to be
            specified. Default is None.
        cluster : compatible to :py:class:``~dask.disributed.LocalCluster`` or
        None
            This dask cluster is used to initialize a
            :py:class:``~dask.distributed.Client``, if no client is specified.
            Default is None.
        """
        self._check_client_cluster(client, cluster)
        if self._validate_client(client):
            self._client = client
            self._cluster = client.cluster
        else:
            self._cluster = cluster
            self._client = Client(cluster)

    def update_state(
            self,
            state: xr.DataArray,
            observations: Union[xr.Dataset, Iterable[xr.Dataset]],
            pseudo_state: xr.DataArray,
            analysis_time: pd.Timestamp
    ) -> xr.DataArray:
        """
        This method updates the state based on given observations and analysis
        time. This method prepares different states, localize these states,
        iterates over state grid points, calculates the ensemble  weights and
        applies these weight to localized state. The calculation of the
        weights is based on PyTorch, while everything else is calculated with
        Numpy / Xarray.

        Parameters
        ----------
        state : :py:class:`xarray.DataArray`
            This state is updated by this assimilation algorithm and given
            ``observation``. This :py:class:`~xarray.DataArray` should have
            four coordinates, which are specified in
            :py:class:`pytassim.state.ModelState`.
        observations : :py:class:`xarray.Dataset` or \
        iterable(:py:class:`xarray.Dataset`)
            These observations are used to update given state. An iterable of
            many :py:class:`xarray.Dataset` can be used to assimilate different
            variables. For the observation state, these observations are
            stacked such that the observation state contains all observations.
        pseudo_state : :py:class:`xarray.DataArray`
            This state is used to generate an observation-equivalent. This
             :py:class:`~xarray.DataArray` should have four coordinates, which
             are specified in :py:class:`pytassim.state.ModelState`.
        analysis_time : :py:class:`datetime.datetime`
            This analysis time determines at which point the state is updated.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            The analysed state based on given state and observations. The
            analysis has same coordinates as given ``state``. If filtering mode
            is on, then the time axis has only one element.
        """
        logger.info('####### {0:s} #######'.format(self._name))
        logger.info('Starting with specific preparation')
        pseudo_obs, obs_state, obs_cov, obs_grid = self._get_states(
            pseudo_state, observations,
        )

        logger.info('Transfering the data to torch')
        pseudo_obs, obs_state, obs_cov = self._states_to_torch(
            pseudo_obs, obs_state, obs_cov
        )
        pseudo_tensor = dask.delayed(torch.zeros(0).to(obs_state))

        logger.info('Normalise perturbations and observations')
        obs_cinv = self._get_chol_inverse(obs_cov)
        normed_perts, normed_obs = self._normalise_obs(pseudo_obs, obs_state,
                                                       obs_cinv)

        logger.info('Chunking and split background state')
        state = state.chunk(
            {'grid': self.chunksize, 'var_name': -1, 'time': -1, 'ensemble': -1}
        )
        state_grid = da.from_array(
            state['grid'].values, chunks=self.chunksize
        )
        chunk_pos = np.concatenate([[0], np.cumsum(state.chunks[-1])])
        state_mean, state_perts = state.state.split_mean_perts()

        logger.info('Scatter data')
        normed_perts, normed_obs, obs_grid = self.client.scatter(
            [normed_perts, normed_obs, obs_grid], broadcast=True
        )

        @dask.delayed
        def slice_data(array_to_slice, min_bound, max_bound):
            sliced_array = array_to_slice[..., min_bound:max_bound]
            return sliced_array

        @dask.delayed
        def to_tensor(array_to_convert, as_tensor):
            converted_tensor = torch.from_numpy(array_to_convert).to(as_tensor)
            return converted_tensor

        @dask.delayed
        def add_mean(perts, mean):
            added_mean = perts + mean.unsqueeze(dim=-2)
            return added_mean

        logger.info('Create analysis')
        analysis_list = []
        for k, pos in enumerate(chunk_pos[1:]):
            loc_perts = dask.delayed(slice_data)(
                state_perts.data, chunk_pos[k], pos
            )
            loc_perts = dask.delayed(to_tensor)(loc_perts, pseudo_tensor)
            loc_grid = dask.delayed(slice_data)(state_grid, chunk_pos[k], pos)
            loc_perts = dask.delayed(self.analyser)(
                loc_perts, normed_perts, normed_obs, loc_grid, obs_grid
            )
            loc_mean = dask.delayed(slice_data)(
                state_mean.data, chunk_pos[k], pos
            )
            loc_mean = dask.delayed(to_tensor)(loc_mean, pseudo_tensor)
            loc_ana = dask.delayed(add_mean)(loc_perts, loc_mean)
            analysis_list.append(loc_ana)

        @dask.delayed
        def cat_numpy(list_to_cat):
            concatenated_list = torch.cat(list_to_cat, dim=-1)
            concatenated_numpy = concatenated_list.numpy()
            return concatenated_numpy

        analysis = dask.delayed(cat_numpy)(analysis_list)
        analysis = analysis.compute()
        analysis = state.copy(data=analysis)
        return analysis


class DistributedLETKFCorr(CorrMixin, DistributedLETKFBase):
    """
    This is a dask-based implementation of the `localized ensemble transform
    Kalman filter` :cite:`hunt_efficient_2007` for correlated observations.
    This dask-based implementation is based on
    :py:class:``~dask.distributed.Client``.

    Parameters
    ----------
    client : :py:class:``~dask.distributed.Client`` or None
        This dask distributed client is used to parallelize the processes.
        Either this client or ``cluster`` has to be
        specified. Default is None. If both, cluster and client, are given, then
        client has priority.
    cluster : compatible to :py:class:``~dask.disributed.LocalCluster`` or None
        This dask cluster is used to initialize a :py:class:``~dask.distributed.
        Client``, if no client is specified.
        Either this cluster or a ``client`` has to be given. Default is None.
    chunksize : int, optional
        The data is splitted up such that every chunk has this number of
        samples. This influences the performance of this distributed version of
        the LETKF. Default is 10.
    localization : obj or None, optional
        This localization is used to localize and constrain observations
        spatially. If this localization is None, no localization is applied such
        it is an inefficient version of the `ensemble transform Kalman filter`.
        Default value is None, indicating no localization at all.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Correlated {0:s}'.format(str(super(DistributedLETKFBase)))

    def __repr__(self):
        return 'Corr{0:s}'.format(repr(super(DistributedLETKFBase)))


class DistributedLETKFUncorr(UnCorrMixin, DistributedLETKFBase):
    """
    This is a dask-based implementation of the `localized ensemble transform
    Kalman filter` :cite:`hunt_efficient_2007` for uncorrelated observations.
    This dask-based implementation is based on
    :py:class:``~dask.distributed.Client``.

    Parameters
    ----------
    chunks : int, optional
        The data is splitted up in this number of chunks.
    localization : obj or None, optional
        This localization is used to localize and constrain observations
        spatially. If this localization is None, no localization is applied such
        it is an inefficient version of the `ensemble transform Kalman filter`.
        Default value is None, indicating no localization at all.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Uncorrelated {0:s}'.format(str(super(DistributedLETKFBase)))

    def __repr__(self):
        return 'Uncorr{0:s}'.format(repr(super(DistributedLETKFBase)))



from .filter import *
from .neural import *
from .variational import *

__all__ = ['ETKFCorr', 'ETKFUncorr', 'LETKFCorr', 'LETKFUncorr',
           'DistributedLETKFCorr', 'DistributedLETKFUncorr',
           'NeuralAssimilation'
           ]



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 18.08.20
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Type, Union, Tuple, Iterable

# External modules
import torch

# Internal modules
from .ketkf_core import KETKFAnalyser
from .etkf_core import CorrMixin, UnCorrMixin
from .etkf import ETKFBase

from pytassim.kernels.base_kernels import BaseKernel
from pytassim.transform.base import BaseTransformer


logger = logging.getLogger(__name__)


__all__ = ['KETKFCorr', 'KETKFUncorr']


class KETKFBase(ETKFBase):
    """
    The base class for the kernelized ensemble transform Kalman filter.
    """
    def __init__(
            self,
            kernel: Type[BaseKernel],
            inf_factor: Union[torch.Tensor, float, torch.nn.Parameter] = 1.0,
            smoother: bool = False, gpu: bool = False,
            pre_transform: Union[None, Iterable[Type[BaseTransformer]]] = None,
            post_transform: Union[None, Iterable[Type[BaseTransformer]]] = None
    ):
        self._analyser = KETKFAnalyser(kernel=kernel, inf_factor=inf_factor)
        super().__init__(inf_factor=inf_factor, smoother=smoother, gpu=gpu,
                         pre_transform=pre_transform,
                         post_transform=post_transform)
        self._name = 'Global Kernel ETKF'

    def __str__(self):
        return 'Kernel ETKF ({0:s}, {1})'.format(str(self.kernel),
                                                 self.inf_factor)

    def __repr__(self):
        return 'KETKF({0:s})'.format(repr(self.kernel))

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self._analyser.inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        self._analyser = KETKFAnalyser(
            kernel=self._analyser.kernel, inf_factor=new_factor
        )

    @property
    def kernel(self) -> Type[BaseKernel]:
        return self._analyser.kernel

    @kernel.setter
    def kernel(self, new_kernel: Type[BaseKernel]):
        self._analyser = KETKFAnalyser(
            kernel=new_kernel, inf_factor=self._analyser.inf_factor
        )


class KETKFCorr(CorrMixin, KETKFBase):
    """
    This is a kernelised verison of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001` for correlated observation covariances.
    This kernelised ensemble Kalman filter is a deterministic filter, where the
    state is globally updated. Ensemble weights are estimated in a reduced
    ensemble space, and then applied to a given state. This kernelised data
    assimilation can be used for non-linear observation operators. The
    observation operator is approximated by the ensemble and a given kernel.
    Furthermore, this implementation allows filtering in time and ensemble
    smoothing, similar to :cite:`hunt_four-dimensional_2004`. This ETKF
    implementation is less efficient for a linear kernel than
    :py:class:`pytassim.assimilation.filter.etkf.ETKFCorr`, which should be
    then used instead.

    Parameters
    ----------
    kernel : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This kernel is used to estimate the ensemble distance matrix.
        If no child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        is used, the kernel should have atleast a :py:func:`forward` method.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Correlated {0:s}'.format(str(super(KETKFBase)))

    def __repr__(self):
        return 'Corr{0:s}'.format(repr(super(KETKFBase)))


class KETKFUncorr(UnCorrMixin, KETKFBase):
    """
    This is a kernelised verison of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001` for uncorrelated observation covariances.
    This kernelised ensemble Kalman filter is a deterministic filter, where the
    state is globally updated. Ensemble weights are estimated in a reduced
    ensemble space, and then applied to a given state. This kernelised data
    assimilation can be used for non-linear observation operators. The
    observation operator is approximated by the ensemble and a given kernel.
    Furthermore, this implementation allows filtering in time and ensemble
    smoothing, similar to :cite:`hunt_four-dimensional_2004`. This ETKF
    implementation is less efficient for a linear kernel than
    :py:class:`pytassim.assimilation.filter.etkf.ETKFUncorr`, which should be
    then used instead.

    Parameters
    ----------
    kernel : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This kernel is used to estimate the ensemble distance matrix.
        If no child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        is used, the kernel should have atleast a :py:func:`forward` method.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    smoother : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Uncorrelated {0:s}'.format(str(super(KETKFBase)))

    def __repr__(self):
        return 'Uncorr{0:s}'.format(repr(super(KETKFBase)))



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc

# External modules
import torch.nn

# Internal modules


logger = logging.getLogger(__name__)


class BaseKernel(torch.nn.Module):
    """
    This kernel is the base kernel for all other kernel objects and used to
    add mathematical operations like `__add__`, `__mul__` and `__power__`.
    New kernels should be a child of this BaseKernel and have to overwrite
    the `forward(self, x, y)` method.
    """
    def __add__(self, other: torch.nn.Module) -> torch.nn.Module:
        return AdditiveKernel(self, other)

    def __mul__(self, other: torch.nn.Module) -> torch.nn.Module:
        return MultiplicativeKernel(self, other)

    def __pow__(self, other: torch.nn.Module) -> torch.nn.Module:
        return PowerKernel(self, other)

    @abc.abstractmethod
    def forward(self, x: torch.tensor, y: torch.tensor) -> torch.tensor:
        pass


class CompKernel(BaseKernel):
    """
    This composition kernel is used as base kernel for all types of
    compositions, like addition or multiplication. This composition adds two
    given kernels as :py:class:`torch.nn.Module`, which might be then used to
    compose a new kernel from the given kernels.

    Parameters
    ----------
    kernel_1 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the first kernel, which is used to compose a given new kernel.
    kernel_2 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the second kernel, which is used to compose a given new kernel.

    """
    def __init__(self, kernel_1: BaseKernel, kernel_2: BaseKernel):
        super().__init__()
        self.add_module('kernel_1', kernel_1)
        self.add_module('kernel_2', kernel_2)


class AdditiveKernel(CompKernel):
    """
    The additive kernel adds the second given kernel to the first given kernel.

    .. math::

       K(x_i, x_j) = K_1(x_i, x_j) + K_2(x_i, x_j)


    Parameters
    ----------
    kernel_1 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the first kernel, which is used to compose a given new kernel.
    kernel_2 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the second kernel, which is used to compose a given new kernel.

    """
    def __str__(self) -> str:
        return "{0:s}+{1:s}".format(str(self.kernel_1), str(self.kernel_2))

    def __repr__(self) -> str:
        return "{0:s}+{1:s}".format(repr(self.kernel_1), repr(self.kernel_2))

    def forward(self, x: torch.tensor, y: torch.tensor) -> torch.tensor:
        return self.kernel_1(x, y) + self.kernel_2(x, y)


class MultiplicativeKernel(CompKernel):
    """
    The multiplicative kernel multiplies the second given kernel with the first
    given kernel.

    .. math::

       K(x_i, x_j) = K_1(x_i, x_j) * K_2(x_i, x_j)


    Parameters
    ----------
    kernel_1 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the first kernel, which is used to compose a given new kernel.
    kernel_2 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the second kernel, which is used to compose a given new kernel.

    """
    def __str__(self) -> str:
        return "{0:s}*{1:s}".format(str(self.kernel_1), str(self.kernel_2))

    def __repr__(self) -> str:
        return "{0:s}*{1:s}".format(repr(self.kernel_1), repr(self.kernel_2))

    def forward(self, x: torch.tensor, y: torch.tensor) -> torch.tensor:
        return self.kernel_1(x, y) * self.kernel_2(x, y)


class PowerKernel(CompKernel):
    """
    The power kernel takes the first given kernel to the power of the second
    given kernel.

    .. math::

       K(x_i, x_j) = {K_1(x_i, x_j)}^{K_2(x_i, x_j)}


    Parameters
    ----------
    kernel_1 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the first kernel, which is used to compose a given new kernel.
    kernel_2 : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This is the second kernel, which is used to compose a given new kernel.

    """
    def __str__(self) -> str:
        return "{0:s}^{1:s}".format(str(self.kernel_1), str(self.kernel_2))

    def __repr__(self) -> str:
        return "{0:s}^{1:s}".format(repr(self.kernel_1), repr(self.kernel_2))


    def forward(self, x: torch.tensor, y: torch.tensor) -> torch.tensor:
        return self.kernel_1(x, y).pow(self.kernel_2(x, y))



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch
import torch.nn

# Internal modules
from .base_kernels import BaseKernel


logger = logging.getLogger(__name__)


__all__ = ['ScaleKernel']


class ScaleKernel(BaseKernel):
    """
    This scale kernel specify a constant value :math:`c`, which can be used to
    specify the output variance of the kernel or a constant shift,


    .. math::

       K(x_i, x_j) = c


    Parameters
    ----------
    scaling : torch.Tensor, optional
        This scaling factor specifies the constant value and is deactivated
        in its default value of 0.

    """
    def __init__(self, scaling: torch.Tensor = torch.tensor(0.)):
        super().__init__()
        self.scaling = scaling

    def __str__(self) -> str:
        return 'ScaleKernel({0})'.format(str(self.scaling))

    def __repr__(self) -> str:
        return repr(self.scaling)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        const_tensor = torch.ones(x.shape[:-1] + (y.shape[-2], ))
        const_tensor = const_tensor * self.scaling
        return const_tensor



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch
import torch.nn

# Internal modules
from .base_kernels import BaseKernel
from .utils import euclidean_dist


logger = logging.getLogger(__name__)


__all__ = ['RationalKernel']


class RationalKernel(BaseKernel):
    """
    The rational kernel is similar to apply radial basis function kernels
    with different lengthscales. This kernel has a basis lengthscale
    :math:`l` and a weighting factor :math:`a`, which weights small-scale to
    large-scale variations (:cite:`duvenaud_automatic_2014`),

    .. math::

       K(x_i, x_j) = (1 + \\frac{{(x_i-x_j)}^2}{2\\,a\\,l^2})^{-a}.


    Parameters
    ----------
    lengthscale : torch.Tensor, optional
        This lengthscale specifies the scale of variations. The
        default value of 1 assumes that the input is already normalized.
    weighting : torch.Tensor, optional
        This weighting factor specifies the relative weighting between
        small-scale to large-scale variations (default=1.)

    """
    def __init__(self, lengthscale: torch.Tensor = torch.tensor(1.),
                 weighting: torch.Tensor = torch.tensor(1.)):
        super().__init__()
        self.lengthscale = lengthscale
        self.weighting = weighting

    def __str__(self) -> str:
        return 'RationalKernel({0}, {1})'.format(
            str(self.lengthscale), str(self.weighting)
        )

    def __repr__(self) -> str:
        return 'Rational({0}, {1})'.format(repr(self.lengthscale),
                                           repr(self.weighting))

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x_scaled = x.div(self.lengthscale)
        y_scaled = y.div(self.lengthscale)
        euc_dist = euclidean_dist(x_scaled, y_scaled)
        norm_factor = 2 * self.weighting
        factor = 1 + euc_dist / norm_factor
        k_mat = factor.pow(-self.weighting)
        return k_mat



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/10/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Type

# External modules
from tqdm import tqdm
import xarray as xr
import numpy as np
import torch

# Internal modules
from pytassim.model.integration.integrator import BaseIntegrator


logger = logging.getLogger(__name__)


def forward_model(
        all_steps: np.ndarray,
        start_point: int,
        start_state: torch.Tensor,
        integrator: Type[BaseIntegrator]
) -> xr.DataArray:
    """
    This forward model can be used to integrate the Lorenz 96 in time.

    Parameters
    ----------
    all_steps : :py:class:`np.ndarray`
        All time steps, which should be estimated.
    start_point : int
        Starting from this point, all steps are written as output.
    start_state : :py:class:`torch.Tensor`
        This start state is used as initial value for the integration.
    integrator : child of :py:class:`pytassim.model.integration.integrator.BaseIntegrator
        This integrator is used to integrate the model.
    """
    output_steps = all_steps[all_steps >= start_point]
    dataset = []
    curr_state = start_state
    for t in tqdm(all_steps, leave=False):
        curr_state = integrator.integrate(curr_state)
        if t in output_steps:
            dataset.append(curr_state)
    stacked_dataset = torch.stack(dataset, dim=-1).numpy()
    stacked_dataset = np.array(stacked_dataset, ndmin=4)
    stacked_dataset = stacked_dataset.transpose(0, 3, 1, 2)
    _, _, ens_size, nr_grid = stacked_dataset.shape
    dataset = xr.DataArray(
        data=stacked_dataset,
        coords={
            'var_name': ['x', ],
            'time': output_steps,
            'ensemble': np.arange(ens_size),
            'grid': np.arange(nr_grid)
        },
        dims=['var_name', 'time', 'ensemble', 'grid']
    )
    return dataset



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch

# Internal modules


logger = logging.getLogger(__name__)


def dot_product(
        x: torch.Tensor,
        y: torch.Tensor
) -> torch.Tensor:
    """
    The linear kernel defined by the dot product between x and y.

    Parameters
    ----------
    x : :py:class:`torch.Tensor` (n_samples_x, n_features)
        The first input to the kernel.
    y : :py:class:`torch.Tensor` (n_samples_y, n_features)
        The second input to the kernel.

    Returns
    -------
    mat : :py:class:`torch.Tensor` (n_samples_x, n_samples_y)
        The dot product between given `x` and `y`.
    """
    mat = torch.einsum('...ij,...kj->...ik', x, y)
    return mat


def distance_matrix(
        x: torch.Tensor,
        y: torch.Tensor,
        norm: float = 2.
) -> torch.Tensor:
    """
    The distance matrix between given tensors defined via the p-norm distance.

    Parameters
    ----------
    x : :py:class:`torch.Tensor` (..., n_samples_x, n_features)
        The first input to the kernel.
    y : :py:class:`torch.Tensor` (..., n_samples_y, n_features)
        The second input to the kernel.
    norm : :py:class:`float`
        The norm of this distance (default=2.0).

    Returns
    -------
    dist : :py:class:`torch.Tensor` (..., n_samples_x, n_samples_y)
        p-norm distance between x and y.
    """
    x_batched = x.unsqueeze(0)
    y_batched = y.unsqueeze(0)
    dist_tensor = torch.cdist(x_batched, y_batched, p=norm)
    dist_squeezed = dist_tensor.squeeze(0)
    return dist_squeezed


def euclidean_dist(
        x: torch.Tensor,
        y: torch.Tensor
) -> torch.Tensor:
    """
    The euclidean distance defined as squared difference between x and y.

    Parameters
    ----------
    x : :py:class:`torch.Tensor` (..., n_samples_x, n_features)
        The first input to the kernel.
    y : :py:class:`torch.Tensor` (..., n_samples_y, n_features)
        The second input to the kernel.

    Returns
    -------
    dist : :py:class:`torch.Tensor` (..., n_samples_x, n_samples_y)
        Euclidean distance between x and y.
    """
    dist = distance_matrix(x, y, norm=2.).pow(2)
    return dist



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 1/25/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import numpy as np

import torch

# Internal modules
from .identity import IdentityOperator


logger = logging.getLogger(__name__)


class BernoulliOperator(IdentityOperator):
    def __init__(self, shift=5, obs_points=None, len_grid=40,
                 random_state=None):
        super().__init__(obs_points=obs_points, len_grid=len_grid,
                         random_state=random_state)
        self.shift = shift

    @staticmethod
    def _np_sigmoid(state):
        return 1. / (1. + np.exp(-state))

    def obs_op(self, in_array, *args, **kwargs):
        obs_state = super().obs_op(in_array, *args, **kwargs)
        obs_state = obs_state - self.shift
        obs_state = self._np_sigmoid(obs_state)
        return obs_state

    def torch_operator(self):
        lin_layer = torch.nn.Linear(self.len_grid, len(self._sel_obs_points),
                                    bias=True)
        for param in lin_layer.parameters():
            param.requires_grad = False

        lin_layer.weight.data = torch.zeros_like(lin_layer.weight.data)
        for k, i in enumerate(self._sel_obs_points):
            lin_layer.weight.data[k, i] = 1.

        lin_layer.bias.data = torch.ones_like(lin_layer.bias.data)
        lin_layer.bias.data *= -self.shift
        operator = torch.nn.Sequential(
            lin_layer, torch.nn.Sigmoid()
        )
        return operator



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch

# Internal modules
from .base_kernels import BaseKernel
from .utils import dot_product


logger = logging.getLogger(__name__)


__all__ = ['PolyKernel']


class PolyKernel(BaseKernel):
    """
    This polynomial kernel specifies a :math:`p` polynomial feature space,
    constructed based on given data with constant value :math:`c`,


    .. math::

       K(x_i, x_j) = ({(x_i)}^{T} x_j + c)^{p}


    Parameters
    ----------
    degree : torch.Tensor, optional
        This specifies the degree of the polynomial. The default of 2
        corresponds to a quadratic kernel.
    const : torch.Tensor, optional
        This constant value is used to shift the center of the kernel. The
        default of 1 activates shifting of the kernel center.

    """
    def __init__(self, degree: torch.Tensor = torch.tensor(2.),
                 const: torch.Tensor = torch.tensor(1.)):
        super().__init__()
        self.degree = degree
        self.const = const

    def __str__(self) -> str:
        return 'PolynomialKernel({0}, {1})'.format(
            str(self.degree), str(self.const)
        )

    def __repr__(self) -> str:
        return 'Polynomial({0}, {1})'.format(repr(self.degree),
                                             repr(self.const))

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        xy = dot_product(x, y)
        k_mat = (xy + self.const).pow(self.degree)
        return k_mat



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/6/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import numpy as np
import torch

# Internal modules


logger = logging.getLogger(__name__)


def dummy_obs_operator(obs_ds, state):
    """
    This dummy observation operator can be used to patch
    :py:meth:`~pytassim.observation.Observation.operator` for testing purpose.
    This function / method select the `x` variable of given ``state`` and
    renames `grid` to `obs_grid_1`. `time` and `obs_grid_1` are further new set
    based on values of given observation instance in self.

    Parameters
    ----------
    obs_ds : :py:class:`~xarray.Dataset`
        This observational dataset is used as basis dataset.
    state : :py:class:`~xarray.DataArray`
        The pseudo observations are created based on this state.

    Returns
    -------
    pseudo_obs : :py:class:`~xarray.DataArray`
        The created pseudo observations based on the given state and this
        observation operator. The last two dimensions are the same
        dimensions as the ``observations`` :py:class:`~xarray.DataArray`
        in set observation subset.
    """
    pseudo_obs = state.sel(var_name='x')
    pseudo_obs = pseudo_obs.rename(grid='obs_grid_1')
    pseudo_obs['time'] = obs_ds.time.values
    pseudo_obs['obs_grid_1'] = obs_ds.obs_grid_1.values
    return pseudo_obs


def dummy_update_state(self, state, observations, pseudo_state, analysis_time):
    """
    This dummy update state can be used to patch
    :py:meth:`~pytassim.assimilation.base.BaseAssimilation.update_state` for
    testing purpose. This dummy update state slices state by nearest
    ``analysis_time``.

    Parameters
    ----------
    self : obj
        This method is patched to this object.
    state : :py:class:`xarray.DataArray`
        This state is used to generate an observation-equivalent. It is
        further updated by this assimilation algorithm and given
        ``observation``. This :py:class:`~xarray.DataArray` should have
        four coordinates, which are specified in
        :py:class:`pytassim.state.ModelState`.
    observations : :py:class:`xarray.Dataset` or \
    iterable(:py:class:`xarray.Dataset`)
        These observations are no used in this dummy method.
    pseudo_state : :py:class:`xarray.DataArray`
        This state is used to generate an observation-equivalent. This
         :py:class:`~xarray.DataArray` should have four coordinates, which
         are specified in :py:class:`pytassim.state.ModelState`.
    analysis_time : :py:class:`datetime.datetime`
        This analysis time determines at which point the state is updated.

    Returns
    -------
    analysis : :py:class:`xarray.DataArray`
        The analysed state based on given state and observations. The
        analysis has same coordinates as given ``state`` except ``time``,
        which should contain only one time step.

    """
    analysis = state.sel(time=analysis_time, method='nearest')
    analysis = analysis.expand_dims('time', axis=1)
    return analysis


def dummy_model(state):
    """
    This model is a dummy model which can be used for testing purpose. This
    dummy model only returns given state without type testing such that it can
    be used for all types of inputs.

    Parameters
    ----------
    state : any
        This `model_state` is returned by this dummy_model.

    Returns
    -------
    derivative : any
        The returned state, same as given ``state``.
    """
    derivative = state
    return derivative


class DummyLocalization(object):
    """
    This localization selects only grid points where `grid_ind` and `obs_grid`
    are the same.
    """
    @staticmethod
    def localize_obs(grid_ind, obs_grid):
        distance = np.abs(obs_grid-grid_ind)
        obs_weights = np.clip(1 - distance / 10, a_min=0, a_max=None)[:, 0]
        use_obs = obs_weights > 0
        return use_obs, obs_weights


def dummy_distance(a, b):
    """
    This calculates a dummy distance as :math:`\\text{abs}(a-b)`.

    Returns
    -------
    distance : any
    """
    distance = np.abs(a-b)
    return distance


class DummyNeuralModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(16, 8, bias=False)

    def forward(self, in_tensor):
        return self.linear(in_tensor)

    @staticmethod
    def assimilate(state, *args, **kwargs):
        return state



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch
import torch.nn

import numpy as np

# Internal modules
from .base_kernels import BaseKernel
from .utils import distance_matrix


logger = logging.getLogger(__name__)


__all__ = ['PeriodicKernel']


class PeriodicKernel(BaseKernel):
    """
    The periodic kernel specifies a periodic process with a given periodicity
    :math:`p` and a lengthscale :math:`l` :cite:`duvenaud_automatic_2014`,


    .. math::

       K(x_i, x_j) = \\exp(-\\frac{2 \\sin^2(\\pi\\mid\\mid x_i - x_j
       \\mid\\mid_1/p)}{l^2})


    Parameters
    ----------
    period : torch.Tensor, optional
        This specifies the periodicity of the kernel (default=1)
    lengthscale : torch.Tensor, optional
        This specifies the length scale, similar to the
        :py:class:`pytassim.kernel.rbf.GaussKernel` (default=1)

    """
    def __init__(self, period: torch.Tensor = torch.tensor(1.),
                 lengthscale: torch.Tensor = torch.tensor(1.)):
        super().__init__()
        self.period = period
        self.lengthscale = lengthscale

    def __str__(self) -> str:
        return 'PeriodicKernel({0}, {1})'.format(
            str(self.period), str(self.lengthscale)
        )

    def __repr__(self) -> str:
        return 'Periodic({0}, {1})'.format(repr(self.period),
                                           repr(self.lengthscale))

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        dist_mat = distance_matrix(x, y, 1.) * np.pi / self.period
        factor = -2 * torch.sin(-dist_mat).pow(2) / (self.lengthscale ** 2)
        return torch.exp(factor)



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 10/4/18
#
# Created for neural_assim
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import numpy as np

import torch

# Internal modules
from pytassim.obs_ops.base_ops import BaseOperator


logger = logging.getLogger(__name__)


class IdentityOperator(BaseOperator):
    def __init__(self, obs_points=None, len_grid=40, random_state=None):
        """
        This linear observation operator is an identity observation operator,
        where observed grid points equal observations.

        Parameters
        ----------
        obs_points : list(int), int or None
            Observed grid points. If this is int, then this number of grid
            points are drawn from grid. If this is a list, observed grid points
            are prescribed. If this is None, all grid points are observed.
            Default is None.
        len_grid : int, optional
            Number of grid points in Lorenz '96 model. Default is 40.
        random_state : :py:class:`numpy.random.RandomState` or None, optional
            This random state can be used for random numbers. Default is None.
        """
        super().__init__(len_grid=len_grid, random_state=random_state)
        self._obs_points = None
        self._sel_obs_points = None
        self.obs_points = obs_points

    @property
    def obs_points(self):
        return self._obs_points

    @obs_points.setter
    def obs_points(self, points):
        if isinstance(points, (int, float)):
            self._sel_obs_points = self.random_state.choice(
                self.len_grid, size=points, replace=False
            )
        elif points is None:
            self._sel_obs_points = np.arange(self.len_grid)
        else:
            self._sel_obs_points = points
        self._obs_points = points

    def obs_op(self, in_array, *args, **kwargs):
        if 'var_name' in in_array.dims:
            in_array = in_array.sel(var_name='x')
        obs_state = in_array.sel(grid=self._sel_obs_points)
        return obs_state

    def torch_operator(self):
        operator = torch.nn.Linear(self.len_grid, len(self._sel_obs_points),
                                   bias=False)
        for param in operator.parameters():
            param.requires_grad = False
        operator.weight.data = torch.zeros_like(operator.weight.data)
        for k, i in enumerate(self._sel_obs_points):
            operator.weight.data[k, i] = 1.
        return operator



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/21/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Iterable

# External modules
import numpy as np
import xarray as xr

# Internal modules
from . import common


logger = logging.getLogger(__name__)


_clm_vcoords = ['levsoi', 'levtot', 'levsno', 'levlak', 'no_vgrid']


def preprocess_clm(
        ds_clm: xr.Dataset,
        assim_vars: Iterable[str]
) -> xr.DataArray:
    """
    Preprocess a given CLM dataset. This dataset is typically created based
    on read-in of `clmoas.clm2.*.nc`files. Only variables specified in
    `assim_vars` are kept.
    """
    sliced_ds = ds_clm[assim_vars]
    ds_gridded = common.create_vgrid(sliced_ds, _clm_vcoords)
    ds_added_no_vgrid = common.add_no_vgrid(
        ds_gridded, _clm_vcoords, ds_gridded['vgrid'].min().values
    )
    ds_interp = _interp_vgrid(ds_added_no_vgrid)
    vertical_remap_dict = {
        'levsoi': 'vgrid', 'levlak': 'vgrid', 'no_vgrid': 'vgrid',
        'levtot': 'vgrid', 'levsno': 'vgrid',
    }

    prepared_ds = common.replace_grid(ds_interp, vertical_remap_dict)
    if 'column' in prepared_ds.dims and 'lat' not in prepared_ds.dims:
        grid_dims = ['column', 'vgrid']
    else:
        grid_dims = ['lat', 'lon', 'vgrid']
    prepared_data = common.ds_to_array(prepared_ds, grid_dims) 
    return prepared_data


def postprocess_clm(
        analysis_data: xr.DataArray,
        ds_clm: xr.Dataset
) -> xr.Dataset:
    """
    This function can be used to post-process CLM data and incorporate
    included variables into given CLM dataset. There are different steps
    included:

    1. Unstack the grid form analysis data and convert the data into a dataset
    2. Iterate through included variables and reindex these variables as they
        are in given CLM dataset
    3. Replace the variables in given CLM dataset with the analysis variables

    Parameters
    ----------
    analysis_data : :py:class:`xarray.DataArray`
        This array represents a valid analysis state, which can be a result of
        any given assimilation algorithm included in this package. In this
        analysis data, the horizontal and vertical grid are stacked together and
        the variables are a dimension within the array.
    ds_clm : :py:class:`xarray.Dataset`
        This CLM dataset is used as source dataset to convert given analysis
        array into a valid CLM dataset. The resulting dataset is a copy of
        this dataset, where the assimilated variables are replaced.

    Returns
    -------
    analysis_ds : :py:class:`xarray.Dataset`
        This analysis dataset is a copy of given CLM dataset with replaced
        variables from given analysis array.
    """
    analysis_ds = common.generic_postprocess(
        analysis_data, ds_clm, vcoords=_clm_vcoords
    )
    logger.info('Finished post-processing of CLM')
    return analysis_ds


def _interp_vgrid(
        ds: xr.Dataset
) -> xr.Dataset:
    """
    Reindexes the `vgrid` coordinate and is used in
    :py:func:`preprocess_clm`.
    """
    avail_vcoords = [c for c in _clm_vcoords if c in ds.dims]
    interp_ds = ds.reindex(
        **{c: ds['vgrid'].values for c in avail_vcoords}, method=None
    )
    return interp_ds



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 02.12.18
#
# Created for torch-assim
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Union

# External modules
import torch

# Internal modules


logger = logging.getLogger(__name__)


def torch_roll(a, shift, axis=0):
    """
    Shift :py:class:`torch.Tensor` along specified axis by `shift`-steps.

    Parameters
    ----------
    a : :py:class:`torch.Tensor`
        This tensor is shifted.
    shift : int
        The tensor is shifted by this number of steps. Positive values shift it
        to the right side, a negative value will shift the tensor to the left
        side.
    axis : int
        The array is shifted along this specified axis. Default is 0, which
        specifies the first axis.

    Returns
    -------
    rolled_tensor : :py:class:`torch.Tensor`
        This tensor was rolled on specified axis by ``shift``-steps. This tensor
        has same type and shape as the input tensor.
    """
    ndims = a.dim()
    left_slice = [slice(None, None)] * ndims
    left_slice[axis] = slice(-shift, None)
    right_slice = [slice(None, None)] * ndims
    right_slice[axis] = slice(None, -shift)
    rolled_tensor = torch.cat((a[left_slice], a[right_slice]), dim=axis)
    return rolled_tensor


class Lorenz96(object):
    """
    The Lorenz '96 :cite:`lorenz_predictability_1996,lorenz_optimal_1998` is a
    grid based dynamical model. In its default
    settings it has a chaotic behaviour. The grid points are wrapped in
    one-dimension such that the first and last grid point are coupled. Only
    surrounding grid points and the forcing F influence the i-th grid point at
    one time step,

    .. math::

       \\frac{dx_{i}}{dt} = (x_{i+1}-x_{i-2})\\,x_{i-1}-x_{i}+F.

    The time unit of this model equals 5 days, a typical synoptical time scale.
    The model includes internal dissipation and external forcing by the linear
    terms, while quadratic terms represent advection.

    Arguments
    ---------
    forcing : float, optional
        The forcing term in the equation. The default forcing of 8 leads to
        typical chaotic behaviour of the atmosphere.
    """
    def __init__(
            self,
            forcing: Union[float, torch.Tensor, torch.nn.Parameter] = 8.
    ):
        self.forcing = forcing

    def __str__(self):
        return 'Lorenz96(F={0})'.format(self.forcing)

    def __repr__(self):
        return 'Lorenz96'

    @staticmethod
    def _calc_advection(state: torch.Tensor) -> torch.Tensor:
        """
        This method calculates the advection term of the Lorenz '96 model. This
        term is given by

        .. math::

           (x_{i+1}-x_{i-2})\\,x_{i-1}.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to calculate the advection term of the Lorenz
            model. The last axis should be the grid axis along which the
            advection is calculated.

        Returns
        -------
        advection : :py:class:`torch.Tensor`
            The calculated advection based on given state. The advection has the
            same type and shape as the input state.
        """
        diff = torch_roll(state, -1, axis=-1) - torch_roll(state, 2, axis=-1)
        advection = diff * torch_roll(state, 1, axis=-1)
        return advection

    def _calc_dissipation(self, state: torch.Tensor) -> torch.Tensor:
        """
        This method calculates the dissipation term. The term is given by

        .. math::

            -x_{i}.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to calculate the dissipation term of the Lorenz
            model.

        Returns
        -------
        dissipation : :py:class:`torch.Tensor`
            The calculated dissipation based on given state. The dissipation has
            same type and shape as the input state.
        """
        dissipation = -state
        return dissipation

    def _calc_forcing(self, state: torch.Tensor) -> torch.Tensor:
        """
        This method calculates the forcing. This returns set forcing, which
        a constant forcing in Lorenz '96 model. This method can be overwritten
        to introduce a coupling between different models. The forcing is
        currently given by

        .. math::

            F.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is currently not used, but can be used to introduce a
            state-dependent forcing.

        Returns
        -------
        forcing : :py:class:`torch.Tensor` or float
            The calculated forcing based on attributes and given state. If it is
            only a float than the same forcing is used at every grid point.
        """
        forcing = self.forcing
        return forcing

    def __call__(self, state: torch.Tensor) -> torch.Tensor:
        """
        Calculates the time-derivative :math:`\\frac{dx_{i}}{dt}` for given
        state.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to estimated the current time-derivative. The
            last axis is supposed to be the grid axis.

        Returns
        -------
        derivative : :py:class:`torch.Tensor`
            The calculated time-derivative based on given state. This
            time-derivative has the same type and shape as given state.
        """
        advection = self._calc_advection(state)
        dissipation = self._calc_dissipation(state)
        forcing = self._calc_forcing(state)

        derivative = advection + dissipation + forcing
        return derivative



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/7/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Type, Union, Tuple, Iterable

# External modules
import xarray as xr
from tqdm import tqdm
import torch

# Internal modules
from .etkf import ETKFBase
from .etkf_core import CorrMixin, UnCorrMixin
from .letkf_core import LETKFAnalyser

from pytassim.localization.localization import BaseLocalization
from pytassim.transform.base import BaseTransformer

logger = logging.getLogger(__name__)


class LETKFBase(ETKFBase):
    """
    The base class for the localised ensemble transform Kalman filter.
    """
    def __init__(
            self,
            localization: Union[None, BaseLocalization] = None,
            inf_factor: Union[torch.Tensor, float, torch.nn.Parameter] = 1.0,
            smoother: bool = False, gpu: bool = False,
            pre_transform: Union[None, Iterable[Type[BaseTransformer]]] = None,
            post_transform: Union[None, Iterable[Type[BaseTransformer]]] = None
    ):
        self._analyser = LETKFAnalyser(localization=localization,
                                       inf_factor=inf_factor)
        super().__init__(inf_factor=inf_factor, smoother=smoother, gpu=gpu,
                         pre_transform=pre_transform,
                         post_transform=post_transform)
        self._analyser = LETKFAnalyser(localization=localization,
                                       inf_factor=inf_factor)
        self._name = 'Sequential LETKF'

    def __str__(self):
        return '{0:s}({1:s}, {2})'.format(self._name, str(self.localization),
                                          self.inf_factor)

    def __repr__(self):
        return 'SeqLETKF({0:s})'.format(repr(self.localization))

    @property
    def localization(self) -> Type[BaseLocalization]:
        return self._analyser.localization

    @localization.setter
    def localization(self, new_locs: Type[BaseLocalization]):
        """
        Sets a new localization.
        """
        self._analyser = LETKFAnalyser(
            localization=new_locs, inf_factor=self.analyser.inf_factor
        )

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self._analyser.inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        """
        Sets a new inflation factor.
        """
        if self.analyser is None:
            localization = None
        else:
            localization = self.analyser.localization
        self._analyser = LETKFAnalyser(
            localization=localization, inf_factor=new_factor
        )


class LETKFCorr(CorrMixin, LETKFBase):
    """
    This is an implementation of the `localized ensemble transform Kalman
    filter` :cite:`hunt_efficient_2007` for correlated observations.
    This is a localized version of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001`. This method iterates independently over each
    grid point in given background state. Given localization instance can be
    used to
    constrain the influence of observations in space. The ensemble weights are
    calculated for every grid point and independently applied to every grid
    point. This implementation follows :cite:`hunt_efficient_2007`, with local
    weight estimation and is implemented in PyTorch. This implementation allows
    filtering in time based on linear propagation assumption
    :cite:`hunt_four-dimensional_2004` and ensemble smoothing.

    Parameters
    ----------
    smoothing : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    localization : obj or None, optional
        This localization is used to localize and constrain observations
        spatially. If this localization is None, no localization is applied such
        it is an inefficient version of the `ensemble transform Kalman filter`.
        Default value is None, indicating no localization at all.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Correlated {0:s}'.format(str(super(LETKFBase)))

    def __repr__(self):
        return 'Corr{0:s}'.format(repr(super(LETKFBase)))


class LETKFUncorr(UnCorrMixin, LETKFBase):
    """
    This is an implementation of the `localized ensemble transform Kalman
    filter` :cite:`hunt_efficient_2007` for uncorrelated observations.
    This is a localized version of the `ensemble transform Kalman filter`
    :cite:`bishop_adaptive_2001`. This method iterates independently over each
    grid point in given background state. Given localization instance can be
    used to
    constrain the influence of observations in space. The ensemble weights are
    calculated for every grid point and independently applied to every grid
    point. This implementation follows :cite:`hunt_efficient_2007`, with local
    weight estimation and is implemented in PyTorch. This implementation allows
    filtering in time based on linear propagation assumption
    :cite:`hunt_four-dimensional_2004` and ensemble smoothing.

    Parameters
    ----------
    smoothing : bool, optional
        Indicates if this filter should be run in smoothing or in filtering
        mode. In smoothing mode, no analysis time is selected from given state
        and the ensemble weights are applied to the whole state. In filtering
        mode, the weights are applied only on selected analysis time. Default
        is False, indicating filtering mode.
    localization : obj or None, optional
        This localization is used to localize and constrain observations
        spatially. If this localization is None, no localization is applied such
        it is an inefficient version of the `ensemble transform Kalman filter`.
        Default value is None, indicating no localization at all.
    inf_factor : float, optional
        Multiplicative inflation factor :math:`\\rho``, which is applied to the
        background precision. An inflation factor greater one increases the
        ensemble spread, while a factor less one decreases the spread. Default
        is 1.0, which is the same as no inflation at all.
    gpu : bool, optional
        Indicator if the weight estimation should be done on either GPU (True)
        or CPU (False): Default is None. For small models, estimation of the
        weights on CPU is faster than on GPU!.
    """
    def __str__(self):
        return 'Uncorrelated {0:s}'.format(str(super(LETKFBase)))

    def __repr__(self):
        return 'Uncorr{0:s}'.format(repr(super(LETKFBase)))




#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/11/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import unittest

# External modules
import torch

# Internal modules


logger = logging.getLogger(__name__)


def if_gpu_decorator(func):     # pragma: no cover
    @unittest.skipIf(not torch.cuda.is_available(), "no GPU")
    def newfunc(self, *args, **kwargs):
        func(self, *args, **kwargs)
    return newfunc



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 12/4/18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch

# Internal modules


logger = logging.getLogger(__name__)


class Lorenz84(object):
    """
    Lorenz '84 :cite:`lorenz_irregularity_1984` is a model with three variables.
    These three variables are
    coupled such that the Hadley circulation is emulated. The `X` variable
    represents the amplitude of the westerly wind current, while the other two
    variables (`Y` and `Z`) show the cosine and sine phase of superimposed
    large-scale eddies. They transport heat from tropics towards the poles.
    Considering these terms, we can construct following time derivative of the
    westerly current

    .. math:: \\frac{dX}{dt} = -Y^2 - Z^2 - aX + aF.


    The amplification of the eddies slows the westerly current, as
    represented in :math:`-Y^2` and :math:`-Z^2`. Mechanical and thermal damping
    decreases the eddy phase and slows the westerly current as shown by the
    linear term :math:`-X`. The damping factor :math:`a` controls the damping
    speed of the westerly current, which is also coupled to the symmetric
    forcing :math:`F`.

    The westerly current is coupled to the phase of the eddies

    .. math::

       \\frac{dY}{dt} &= XY - bXZ - Y + G,

       \\frac{dZ}{dt} &= bXY + XZ - Z.

    The time-derivative of the cosine :math:`\\frac{dY}{dt}` and sine
    :math:`\\frac{dZ}{dt}` phase of eddies is influenced by the amplification of
    eddies caused by interactions with the westerly current :math:`X`. The
    factor :math:`b` in front shows eddy displacement due to advection by the
    current. The asymmetric forcing :math:`G` increases the cosine phase of
    eddies, which is then propagated to westerly current and sine phase via the
    coupling of the three variables.

    The typical time unit of this model is 5 days. The default values of the
    arguments are set such that the model represent typical meteorological
    instability :cite:`lorenz_irregularity_1984`.

    Arguments
    ---------
    damp_factor : float, optional
        The damping factor :math:`a` of the westerly current. This factor
        controls the damping of the current and the influence of thermal
        forcing. Default value is 0.25 such that the current is slower damped
        than the eddies.
    dis_factor : float, optional
        The displacement factor :math:`b`, controlling the displacement of the
        eddies by the westerly current. Default value is 4.0, leading to more
        rapid displacement of the eddies than amplification.
    symm_forcing : float, optional
        The symmetric thermal forcing :math:`F` of the westerly current. This
        forcing increases the amplification of the westerly current. If the
        variables were not coupled the amplification would converge to this
        value. Default value is 8.
    asymm_forcing : float, optional
        The asymmetric thermal forcing :math:`G` of the eddies. This forcing
        influences the eddies. :math:`Y` would converge to this value if the
        variables are not coupled by other terms. Default value is 1.
    """
    def __init__(
            self,
            damp_factor: float = 0.25,
            dis_factor: float = 4.0,
            symm_forcing: float = 8.0,
            asymm_forcing: float = 1.0
    ):
        self.dis_factor = dis_factor
        self.damp_factor = damp_factor
        self.symm_forcing = symm_forcing
        self.asymm_forcing = asymm_forcing

    def __str__(self) -> str:
        return 'Lorenz84({0}, {1}, {2}, {3})'.format(
            self.damp_factor, self.dis_factor, self.symm_forcing,
            self.asymm_forcing
        )

    def __repr__(self) -> str:
        return 'Lorenz84'

    def _calc_westerly(self, state: torch.Tensor) -> torch.Tensor:
        """
        Calculate the amplification of the westerly current, including coupling
        effects, damping and forcing

        .. math:: \\frac{dX}{dt} = -Y^2 - Z^2 - aX + aF.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to estimated the amplification. The
            last axis is supposed to be the variable axis and should have a
            size of three.

        Returns
        -------
        amp : :py:class:`torch.Tensor`
            The estimated amplification of the westerly current, depending on
            given state, set damping factor and symmetric forcing.
        """
        coupling = -state[..., 1] ** 2 - state[..., 2] ** 2
        damping = self.damp_factor * state[..., 0]
        forcing = self.damp_factor * self.symm_forcing
        amp = coupling - damping + forcing
        return amp

    def _calc_cosine_phase(self, state: torch.Tensor) -> torch.Tensor:
        """
        Calculates the change in the cosine phase of the eddies, including
        amplification and displacement caused by westerly current, damping and
        thermal forcing

        .. math:: \\frac{dY}{dt} &= XY - bXZ - Y + G.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to estimated the phase change. The
            last axis is supposed to be the variable axis and should have a
            size of three.

        Returns
        -------
        phase_change : :py:class:`torch.Tensor`
            The estimated cosine phase change of the eddies, depending on given
            state, set displacement factor and asymmetric forcing.
        """
        amp = state[..., 0] * state[..., 1]
        displace = -self.dis_factor * state[..., 0] * state[..., 2]
        damping = state[..., 1]
        forcing = self.asymm_forcing
        phase_change = amp + displace - damping + forcing
        return phase_change

    def _calc_sine_phase(self, state: torch.Tensor) -> torch.Tensor:
        """
        Calculates the change in the sine phase of the eddies, including
        amplification and displacement caused by westerly current and damping.

        .. math:: \\frac{dZ}{dt} &= bXY + XZ - Z.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to estimated the phase change. The
            last axis is supposed to be the variable axis and should have a
            size of three.

        Returns
        -------
        phase_change : :py:class:`torch.Tensor`
            The estimated sine phase change of the eddies, depending on given
            state and set displacement factor.
        """

        amp = state[..., 0] * state[..., 2]
        displace = self.dis_factor * state[..., 0] * state[..., 1]
        damping = state[..., 2]
        phase_change = amp + displace - damping
        return phase_change

    def __call__(self, state: torch.Tensor) -> torch.Tensor:
        """
        Estimate the time-derivative of this model. The time-derivative is
        estimated based on amplification of the westerly current, and cosine and
        sine phase change of the eddies.

        Parameters
        ----------
        state : :py:class:`torch.Tensor`
            This state is used to estimated the current time-derivative. The
            last axis is supposed to be the variable axis and should have a size
            of three.

        Returns
        -------
        derivative : :py:class:`torch.Tensor`
            The calculated time-derivative based on given state. This
            time-derivative has the same type and shape as given state.
        """
        westerly_amp = self._calc_westerly(state)
        cosine_change = self._calc_cosine_phase(state)
        sine_change = self._calc_sine_phase(state)

        derivative = torch.stack([westerly_amp, cosine_change, sine_change],
                                 dim=-1)
        return derivative



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 14.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc
from typing import Any, Tuple

# External modules
import numpy as np

# Internal modules


logger = logging.getLogger(__name__)


class BaseLocalization(object):
    """
    This base localization should be used if a localization algorithm is
    implemented.
    """
    @abc.abstractmethod
    def localize_cov(self):
        """
        This method creates a localized covariance based on given parameters.
        """
        pass

    @abc.abstractmethod
    def localize_obs(
            self,
            grid_ind: Any,
            obs_grid: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method creates weights for observations based on given grid index
        and observation grid.

        Parameters
        ----------
        grid_ind : any
            This parameter indicates the current grid index.
        obs_grid : :py:class:`np.ndarray`
            This observation grid is used to estimate a spatial distance to
            given grid index.

        Returns
        -------
        use_obs : :py:class:`np.ndarray`, dtype=bool
            This is a boolean array indicating if the `i`-th observation should
            be used. This array is calculated based on estimated observation
            weights.
        obs_weights : :py:class:`np.ndarray`, dtype=float
            The estimated observation weights. These weights can be used to
            weight observations.
        """
        pass



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 20.08.20
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Type, Union, Iterable, Tuple

# External modules
import torch

# Internal modules
from .etkf_core import ETKFWeightsModule, ETKFAnalyser
from ..utils import evd, rev_evd
from pytassim.kernels import LinearKernel
from pytassim.kernels.base_kernels import BaseKernel


logger = logging.getLogger(__name__)


class KETKFWeightsModule(ETKFWeightsModule):
    """
    This weights module estimate ensemble weights based on the kernelized
    ensemble transform Kalman filter and represents the inner core of the KETKF.

    Parameters
    ----------
    kernel : child of :py:class:`pytassim.kernels.base_kernels.BaseKernel`
        This kernel is used to estimate the ensemble weights.
    inf_factor : torch.Tensor or torch.nn.Parameter, optional
        The prior covariance is inflated with this inflation factor. This
        inflation factor is also a type of l2-regularization for the Gaussian
        processes and specifies the uncertainty of the prior ensemble weights.
    """
    def __init__(
            self,
            kernel: Type[BaseKernel],
            inf_factor: Union[float, torch.Tensor, torch.nn.Parameter]
                      = torch.tensor(1.0)
    ):
        super().__init__(inf_factor)
        self.add_module('kernel', kernel)

    def __str__(self):
        return 'KETKFWeightsModule({0:s}, {1})'.format(str(self.kernel),
                                                       self.inf_factor)

    def __repr__(self):
        return 'KETKFWeights({0:s})'.format(repr(self.kernel))

    def _apply_kernel(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Applies set kernel on given `x` and `y`.
        """
        return self.kernel(x, y)

    def _estimate_weights(
            self,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Estimate the ensemble weights with set kernel and inflation factor, and
        given perturbations and observations.
        """
        ens_size = normed_perts.shape[0]
        reg_value = (ens_size-1) / self._inf_factor

        k_perts = self._apply_kernel(normed_perts, normed_perts)
        k_part_mean = torch.mean(k_perts, dim=-1, keepdim=True)
        k_part_mean = k_part_mean - torch.mean(k_part_mean, dim=-2,
                                               keepdim=True)
        k_perts_centered = k_perts - torch.mean(k_perts, dim=-2,
                                                keepdim=True) - k_part_mean

        evals, evects, evals_inv = evd(k_perts_centered, reg_value)
        cov_analysed = rev_evd(evals_inv, evects)

        k_obs = self._apply_kernel(normed_perts, normed_obs)
        k_obs_centered = k_obs - torch.mean(k_obs, dim=-2, keepdim=True) - \
                         k_part_mean
        w_mean = torch.einsum(
            '...ij,...jk->...ik', cov_analysed, k_obs_centered
        )

        square_root_einv = ((ens_size - 1) * evals_inv).sqrt()
        w_perts = rev_evd(square_root_einv, evects)
        return w_mean, w_perts, cov_analysed


class KETKFAnalyser(ETKFAnalyser):
    """
    Analyser to get analysis perturbations based on set kernel, and given
    background perturbations and normalized observational quantities.
    """
    def __init__(
            self,
            kernel: Type[BaseKernel],
            inf_factor: Union[float, torch.Tensor, torch.nn.Parameter]
                      = torch.tensor(1.0)
    ):
        self.gen_weights = KETKFWeightsModule(kernel=kernel,
                                              inf_factor=inf_factor)
        super().__init__(inf_factor=inf_factor)
        self.kernel = kernel

    def __str__(self) -> str:
        return 'KETKFAnalyser({0:s}, {1})'.format(str(self.kernel),
                                                  self.inf_factor)

    def __repr__(self) -> str:
        return 'KETKFAnalyser({0:s})'.format(repr(self.kernel))

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self.gen_weights.inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        """
        Sets a new inflation factor.
        """
        self.gen_weights = KETKFWeightsModule(
            kernel=self.gen_weights.kernel, inf_factor=new_factor
        )

    @property
    def kernel(self) -> Type[BaseKernel]:
        return self.gen_weights.kernel

    @kernel.setter
    def kernel(self, new_kernel: Type[BaseKernel]):
        """
        Sets a new kernel.
        """
        self.gen_weights = KETKFWeightsModule(
            kernel=new_kernel, inf_factor=self.gen_weights.inf_factor
        )



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 13.08.20
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Union, Tuple, Any

# External modules
import torch.jit
import numpy as np

# Internal modules
from ..utils import grid_to_array
from .etkf_core import ETKFAnalyser, ETKFWeightsModule

from pytassim.localization import BaseLocalization


logger = logging.getLogger(__name__)


class LETKFAnalyser(ETKFAnalyser):
    """
    This analyser uses the etkf weight module and wraps in an outer loop the
    localisation.
    """
    def __init__(
            self,
            localization: Union[None, BaseLocalization] = None,
            inf_factor: Union[torch.Tensor, float, torch.nn.Parameter] = 1.0,
    ):
        self._gen_weights = None
        self._inf_factor = None
        self.localization = localization
        super().__init__(inf_factor)

    def __str__(self) -> str:
        return 'LETKFAnalyser({0:s}, {1})'.format(str(self.localization),
                                                  self.inf_factor)

    def __repr__(self) -> str:
        return 'LETKFAnalyser({0:s})'.format(repr(self.localization))

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self._inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        self._inf_factor = new_factor
        self.gen_weights = ETKFWeightsModule(new_factor)

    @property
    def gen_weights(self) -> ETKFWeightsModule:
        return self._gen_weights

    @gen_weights.setter
    def gen_weights(self, new_module: ETKFWeightsModule):
        if new_module is None:
            self._gen_weights = None
        elif isinstance(new_module, ETKFWeightsModule):
            self._gen_weights = new_module
        else:
            raise TypeError('Given weights module is not a valid '
                            '`ETKFWeightsModule or None!')

    def _localise_obs(
            self,
            grid_point: Any,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor,
            obs_grid: np.ndarray
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Localises given observational quantities based on set localization
        and given grid points.
        """
        if self.localization is None:
            return normed_perts, normed_obs
        else:
            use_obs, obs_weights = self.localization.localize_obs(
                grid_point, obs_grid
            )
            obs_weights = torch.as_tensor(
                obs_weights[use_obs], dtype=normed_perts.dtype
            ).sqrt()
            normed_perts = normed_perts[..., use_obs] * obs_weights
            normed_obs = normed_obs[..., use_obs] * obs_weights
            return normed_perts, normed_obs

    def get_analysis_perts(
            self,
            state_perts: torch.Tensor,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor,
            state_grid: np.ndarray,
            obs_grid: np.ndarray
    ) -> torch.Tensor:
        """
        Estimates analysis perturbations based on set localization and given
        quantities.
        """
        self._gen_weights = torch.jit.script(self._gen_weights)
        grid_index = grid_to_array(state_grid)
        analysis_perts = []
        for ind, grid_point in enumerate(grid_index):
            loc_perts, loc_obs = self._localise_obs(
                grid_point, normed_perts, normed_obs, obs_grid
            )
            loc_analysis_perts = super().get_analysis_perts(
                state_perts[..., [ind]], loc_perts, loc_obs, None, None
            )
            analysis_perts.append(loc_analysis_perts)
        analysis_perts = torch.cat(analysis_perts, dim=-1)
        return analysis_perts



from .dummy import *
from .decorators import *


__all__ = ['dummy_update_state', 'dummy_obs_operator', 'dummy_model',
           'DummyLocalization', 'dummy_distance', 'DummyNeuralModule',
           'if_gpu_decorator']



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 26.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import copy

# External modules

# Internal modules
from ..base import BaseAssimilation


logger = logging.getLogger(__name__)


class FilterAssimilation(BaseAssimilation):
    """
    FilterAssimilation is a base class for all assimilation algorithms, which
    are based on a filtering approach e.g. Ensemble Kalman filters or particle
    filters. This class extends
    :py:class:`~pytassim.assimilation.base.BaseAssimilation` with different
    methods and attributes for a simplified filtering approach.
    """
    def __str__(self):
        return 'FilterAssimilation'

    def __repr__(self):
        return 'FilterAssimilation'



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 2/14/19
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2019}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
from typing import Union, Tuple, Iterable

# External modules
import numpy as np
import torch
import torch.sparse
import scipy.linalg
import xarray as xr
import dask.array as da

# Internal modules
from ..utils import evd, rev_evd


logger = logging.getLogger(__name__)


class ETKFWeightsModule(torch.nn.Module):
    """
    Module to create ETKF weights based on PyTorch.
    This module estimates weight statistics with given perturbations and
    observations.
    """
    def __init__(
            self,
            inf_factor: Union[float, torch.Tensor, torch.nn.Parameter] = 1.0
    ):
        super().__init__()
        self._inf_factor = None
        self.inf_factor = inf_factor

    def __str__(self) -> str:
        return 'ETKFWeightsModule({0})'.format(self.inf_factor)

    def __repr__(self) -> str:
        return 'ETKFWeightsModule'

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self._inf_factor

    @inf_factor.setter
    def inf_factor(
            self, new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        """
        Sets a new inflation factor.
        """
        if isinstance(new_factor, (torch.Tensor, torch.nn.Parameter)):
            self._inf_factor = new_factor
        else:
            self._inf_factor = torch.tensor(new_factor)

    @staticmethod
    def _test_sizes(normed_perts: torch.Tensor, normed_obs: torch.Tensor):
        """
        Tests if sizes between perturbations and observations match.
        """
        if normed_perts.shape[-1] != normed_obs.shape[-1]:
            raise ValueError(
                'Observational size between ensemble ({0:d}) and observations '
                '({1:d}) do not match!'.format(
                    normed_perts.shape[-1], normed_obs.shape[-1]
                )
            )
        if normed_perts.shape[:-2] != normed_obs.shape[:-2]:
            raise ValueError(
                'Batch sizes between ensemble {0} and observations {1} do not '
                'match!'.format(
                    tuple(normed_perts.shape[:-2]), tuple(normed_obs.shape[:-2])
                )
            )

    @staticmethod
    def _apply_kernel(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Apply set kernel matrix, here the dot product, to given tensors.
        """
        k_mat = torch.einsum('...ij,...kj->...ik', x, y)
        return k_mat

    def _get_prior_weights(
            self,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Get prior weights. The perturbations and covariance matrix are
        already inflated by set inflation factor.
        """
        ens_size = normed_perts.shape[-2]
        prior_mean = torch.zeros(normed_perts.shape[:-1]+(1,)).to(normed_perts)
        prior_eye = torch.ones(normed_perts.shape[:-1]).to(normed_perts)
        prior_eye = torch.diag_embed(prior_eye)
        prior_cov = self._inf_factor / (ens_size-1) * prior_eye
        prior_perts = self._inf_factor.sqrt() * prior_eye
        return prior_mean, prior_perts, prior_cov

    def _estimate_weights(
            self,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Estimates the weights with set inflation factor, _apply_kernel method
        and given data.
        """
        ens_size = normed_perts.shape[-2]
        reg_value = (ens_size-1) / self._inf_factor
        kernel_perts = self._apply_kernel(normed_perts, normed_perts)
        evals, evects, evals_inv = evd(kernel_perts, reg_value)
        cov_analysed = rev_evd(evals_inv, evects)

        kernel_obs = self._apply_kernel(normed_perts, normed_obs)
        w_mean = torch.einsum('...ij,...jk->...ik', cov_analysed, kernel_obs,)

        square_root_einv = ((ens_size - 1) * evals_inv).sqrt()
        w_perts = rev_evd(square_root_einv, evects)
        return w_mean, w_perts, cov_analysed

    def forward(
            self,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Get the ensemble weights for given inflation factor, _apply_kernel
        method and data.
        If the perturbations and observations are empty, the inflated prior
        weights are returned.
        """
        self._test_sizes(normed_perts, normed_obs)
        if normed_perts.shape[-1] == 0:
            w_mean, w_perts, cov_analysed = self._get_prior_weights(
                normed_perts, normed_obs
            )
        else:
            w_mean, w_perts, cov_analysed = self._estimate_weights(
                normed_perts, normed_obs
            )
        weights = w_mean + w_perts
        return weights, w_mean, w_perts, cov_analysed


class ETKFAnalyser(object):
    """
    Analyser to get analysis perturbations based on given background
    perturbations and normalized observational quantities.
    """
    def __init__(
            self,
            inf_factor: Union[float, torch.Tensor, torch.nn.Parameter] = 1.0
    ):
        self.inf_factor = inf_factor

    def __str__(self) -> str:
        return 'ETKFAnalyser({0})'.format(self.inf_factor)

    def __repr__(self) -> str:
        return 'ETKFAnalyser'

    @property
    def inf_factor(self) -> Union[float, torch.Tensor, torch.nn.Parameter]:
        return self.gen_weights.inf_factor

    @inf_factor.setter
    def inf_factor(
            self,
            new_factor: Union[float, torch.Tensor, torch.nn.Parameter]
    ):
        """
        Sets a new inflation factor.
        """
        self.gen_weights = ETKFWeightsModule(new_factor)

    @staticmethod
    def _weights_matmul(
            perts: torch.Tensor,
            weights: torch.Tensor
    ) -> torch.Tensor:
        """
        Multiply given ensemble perturbations with ensemble weights.
        """
        ana_perts = torch.einsum('...ig,ij->...jg', perts, weights)
        return ana_perts

    def get_analysis_perts(
            self,
            state_perts: torch.Tensor,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor,
            state_grid: np.ndarray,
            obs_grid: np.ndarray
    ) -> torch.Tensor:
        """
        Estimate the analysis perturbations with given data, set inflation
        factor and kernel.
        """
        weights = self.gen_weights(normed_perts, normed_obs)[0]
        weights = weights.detach()
        ana_perts = self._weights_matmul(state_perts, weights)
        return ana_perts

    def __call__(
            self,
            state_perts: torch.Tensor,
            normed_perts: torch.Tensor,
            normed_obs: torch.Tensor,
            state_grid: np.ndarray,
            obs_grid: np.ndarray
    ) -> torch.Tensor:
        return self.get_analysis_perts(state_perts, normed_perts, normed_obs,
                                       state_grid, obs_grid)


class CorrMixin(object):
    _correlated = True

    @staticmethod
    def _get_obs_cov(observations: Iterable[xr.Dataset]) -> np.ndarray:
        """
        Get the observational covariance from given observations.
        """
        cov_stacked_list = []
        for obs in observations:
            len_time = len(obs.time)
            stacked_cov = [obs['covariance'].values] * len_time
            stacked_cov = scipy.linalg.block_diag(*stacked_cov)
            cov_stacked_list.append(stacked_cov)
        obs_cov = scipy.linalg.block_diag(*cov_stacked_list)
        return obs_cov

    @staticmethod
    def _get_chol_inverse(cov: torch.Tensor) -> torch.Tensor:
        """
        Decomposes given covariance with cholesky decomposition and returns the
        inverse of the cholesky decomposition.
        """
        chol_decomp = torch.cholesky(cov)
        chol_inv = chol_decomp.inverse()
        return chol_inv

    @staticmethod
    def _mul_cinv(state: torch.Tensor, cinv: torch.Tensor) -> torch.Tensor:
        """
        Multiplies given tensor with given inverse of the cholesky decomposed
        covariance matrix.
        """
        normed_state = torch.mm(state, cinv)
        return normed_state


class UnCorrMixin(object):
    _correlated = False

    @staticmethod
    def _get_obs_cov(observations: Iterable[xr.Dataset]) -> np.ndarray:
        """
        Get the observational covariance from given observations.
        """
        cov_stacked_list = []
        for obs in observations:
            len_time = len(obs.time)
            stacked_cov = [obs['covariance'].values] * len_time
            stacked_cov = np.concatenate(stacked_cov)
            cov_stacked_list.append(stacked_cov)
        obs_cov = np.concatenate(cov_stacked_list)
        return obs_cov

    @staticmethod
    def _get_chol_inverse(cov: torch.Tensor) -> torch.Tensor:
        """
        Decomposes given covariance with cholesky decomposition and returns the
        inverse of the cholesky decomposition.
        """
        chol_decomp = cov.sqrt()
        chol_inv = 1 / chol_decomp
        return chol_inv

    @staticmethod
    def _mul_cinv(state: torch.Tensor, cinv: torch.Tensor) -> torch.Tensor:
        """
        Multiplies given tensor with given inverse of the cholesky decomposed
        covariance matrix.
        """
        normed_state = state * cinv
        return normed_state



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 03.08.20
#
# Created for 20_kernel_etkf
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2020}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging

# External modules
import torch.nn

# Internal modules
from .base_kernels import BaseKernel
from .utils import dot_product


logger = logging.getLogger(__name__)


__all__ = ['ModuleKernel']


class ModuleKernel(BaseKernel):
    """
    This module kernel can be used to specify an explicit feature
    transformation process. This kernel takes a :py:class:`torch.nn.Module`
    :math:`\\phi`, which is then applied to the given data. The result is
    combined with a linear kernel as follows,


    .. math::

       K(x_i, x_j) = {(\\phi(x_i))}^{T} \\phi(x_y).

    This kernel can be used to implement random fourier features
    :cite:`rahimi_random_2008`, instances of the Nystrom method
    :cite:`drineas_nystrom_2005` or feature extraction with neural networks.

    Parameters
    ----------
    transform_module : :py:class:`torch.nn.Module`
        This transform module is used to transform given data into a new
        feature space.

    """
    def __init__(self, transform_module: torch.nn.Module):
        super().__init__()
        self.add_module('transform', transform_module)

    def __str__(self) -> str:
        return 'ModuleKernel({0})'.format(str(self.transform))

    def __repr__(self) -> str:
        return 'ModuleKernel'

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        x_net = self.transform(x)
        y_net = self.transform(y)
        k_mat = dot_product(x_net, y_net)
        return k_mat



from .diag import DiagKernel
from .linear import LinearKernel
from .module_kernel import ModuleKernel
from .orn_uhl import OrnsteinUhlenbeckKernel
from .periodic import PeriodicKernel
from .polynomial import PolyKernel
from .rational import RationalKernel
from .rbf import RBFKernel, GaussKernel
from .scale import ScaleKernel
from .tanh import TanhKernel



#!/bin/env python
# -*- coding: utf-8 -*-
#
# Created on 14.03.18
#
# Created for torch-assimilate
#
# @author: Tobias Sebastian Finn, tobias.sebastian.finn@uni-hamburg.de
#
#    Copyright (C) {2018}  {Tobias Sebastian Finn}
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# System modules
import logging
import abc

# External modules

# Internal modules


logger = logging.getLogger(__name__)


class BaseTransformer(object):
    """
    The transformer can be used to transform data before (pre) or after (post)
    assimilation. This can be either some type of covariance inflation,
    normalization or something completely different.
    """
    @abc.abstractmethod
    def pre(self, background, observations, first_guess):
        """
        This transform method is called before the assimilation and can change
        input ``first_guess`` and ``observations``.
        Parameters
        ----------
        background : :py:class:`xarray.DataArray`
            This background array is manipulated by this method and should be
            a valid state array.
        observations : iterable(:py:class:`xarray.Dataset`
            These observations are manipulated by this method and should be
            valid observation datasets with observations and covariance as
            array.
        first_guess : :py:class:`xarray.DataArray`
            This first guess array is manipulated by this method and should be
            a valid state array.

        Returns
        -------
        background : :py:class:`xarray.DataArray`
            This background array was manipulated by this method and should be
            a valid state array.
        observations : iterable(:py:class:`xarray.Dataset`
            These observations were manipulated by this method and are
            valid observation datasets with observations and covariance as
            array.
        first_guess : :py:class:`xarray.DataArray`
            This first guess field was manipulated by this method and is a valid
            state array.
        """
        pass

    @abc.abstractmethod
    def post(self, analysis, background, observations, first_guess):
        """
        This transform method is called after the assimilation and can change
        input ``analysis``. Additional ``first_guess`` and ``observations`` can
        be used to manipulate the analysis.

        Parameters
        ----------
        analysis : :py:class:`xarray.DataArray`
            This analysis array is manipulated by this method and should be
            a valid state array.
        background : :py:class:`xarray.DataArray`
            This background array can be used to manipulate given analysis and
            should be a valid state array.
        observations : iterable(:py:class:`xarray.Dataset`
            These observations can be used to manipulate the analysis and should
            be valid observation datasets with observations and covariance as
            array.
        first_guess : :py:class:`xarray.DataArray`
            This first guess array can be used to manipulate given analysis and
            should be a valid state array.

        Returns
        -------
        analysis : :py:class:`xarray.DataArray`
            This analysis array was manipulated by this method and is a valid
            state array.
        """
        pass
